#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) type-checking**
unit tests.

This submodule unit tests the subset of the public API of the public
:mod:`beartype.door` subpackage that pertains to type-checking.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytmark import ignore_warnings

# ....................{ TESTS ~ warnings                   }....................
#FIXME *INCREDIBLY FRAGILE.* This test has unintended ordering issues with other
#tests due to memoization. If another test runs first and passes the
#is_bearable() tester these type hints and thus emits these warnings, this test
#will re-pass these same hints to that tester, *WHICH WILL FAIL TO EMIT THESE
#WARNINGS DUE TO MEMOIZATION.*
#
#Ergo, the only safe means of performing this test is to ensure that this test
#internally clears all relevant beartype caches *BEFORE* performing any work. To
#the best of our understanding, the only relevant beartype cache is that of our
#core beartype._check.code.codemain.make_check_expr() function. Let us consider
#how to sanitize this, please. Hmm... Actually, we're pretty sure there are more
#than one relevant caches. This is probably extremely non-trivial. The only
#other alternative would be for the @callable_cached decorator to append each
#cache it creates to a giant global private list somewhere, which this unit test
#would then iterate over and clear each item of. Yeah. *sigh*
#
#Incidentally, this also explains why the iter_hints_piths_meta() closure fails.
#Why? Because that closure internally iterates *TWICE* over each hint. Clearly,
#the first use of each deprecated hint emits a deprecation warning, which is
#successfully caught; the second use then fails to do so, due to memoization.
#
#For the moment, we simply forcefully define this test *BEFORE* all other tests
#calling the is_bearable() tester. This works... for now, but will absolutely
#stop working when an earlier run test accidentally calls that tester. If you
#are reading this, that probably already happened, huh? I feel your pain.
#Really, I do. I did nothing about it... but I feel it, nonetheless.
def test_door_is_bearable_warnings(hints_meta) -> None:
    '''
    Test the :class:`beartype.door.is_bearable` tester function with respect
    to non-fatal warnings emitted by that tester when passed various problematic
    (e.g., deprecated) type hints.

    Note that this test *cannot* be folded into the comparable
    :func:`.test_door_is_bearable` test. Why? Because that test (and almost all
    other tests testing type hints) iterates over type hints by calling the
    session-scoped ``iter_hints_piths_meta()`` closure iterator. For unknown
    reasons, :mod:`pytest` fails to capture deprecation warnings emitted from
    that iterator. This is almost certainly a :mod:`pytest` issue. However,
    reporting this issue would require reducing this issue to a minimal
    reproducible working example -- which appears to be infeasible. In short,
    this test manually iterates over type hints as an acceptable (albeit
    annoying) alternative that allows deprecation warnings to be captured.

    Parameters
    ----------
    hints_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintNonpepMetadata]
        List of PEP-agnostic type hint metadata describing sample PEP-agnostic
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype.door import is_bearable
    from pytest import warns
    from warnings import simplefilter

    # ....................{ SETUP                          }....................
    # Force pytest to temporarily allow deprecation warnings to be caught by the
    # warns() context manager for the duration of this test. By default, pytest
    # simply "passes through" all deprecation warnings for subsequent reporting
    # if tests otherwise successfully pass. Deprecation warnings include:
    # * "DeprecationWarning".
    # * "FutureWarning".
    # * "PendingDeprecationWarning".
    simplefilter('always')

    # ..................{ PASS                               }..................
    # For each predefined type hint and associated metadata...
    for hint_meta in hints_meta:
        # If it is *NOT* the case that...
        if not (
            # This hint is currently supported *AND*...
            hint_meta.is_supported and
            # This is type-checkable against at least one object.
            hint_meta.piths_meta
        # Then this hint is ignorable. Silently continue to the next.
        ):
            continue
        # Else, this hint is currently supported.

        # Type hint to be type-checked.
        hint = hint_meta.hint

        # Beartype dataclass configuring this type-check.
        conf = hint_meta.conf

        # Object to be type-checked against this hint, arbitrarily selected from
        # the iterable of all such objects supplied with this hint. By the above
        # validation, this is guaranteed to be non-empty.
        pith = hint_meta.piths_meta[0]

        # If this tester is expected to emit a warning for this hint...
        if hint_meta.warning_type is not None:
            # print(f'hint_meta: {hint_meta}\n')

            # Call this tester under a context manager asserting this tester to
            # emit the expected warning.
            with warns(hint_meta.warning_type):
                is_bearable(pith, hint, conf=conf)
            # print(f'Deprecated type hint {repr(hint)} warned!')
        # Else, this tester is expected to emit *NO* warning for this hint. In
        # this case, call this tester outside of such a context manager.
        else:
            is_bearable(pith, hint, conf=conf)
        # is_bearable(pith, hint, conf=conf)

# ....................{ TESTS ~ raisers                    }....................
# Prevent pytest from capturing and displaying all expected non-fatal
# beartype-specific warnings emitted by this test. Urgh!
@ignore_warnings(DeprecationWarning)
def test_door_die_if_unbearable(iter_hints_piths_meta) -> None:
    '''
    Test the :class:`beartype.door.die_if_unbearable` raiser function.

    Parameters
    ----------
    iter_hints_piths_meta : Callable[[], Iterable[beartype_test.a00_unit.data.hint.util.data_hintmetautil.HintPithMetadata]]
        Factory function creating and returning a generator iteratively yielding
        ``HintPithMetadata`` instances, each describing a sample type hint
        exercising an edge case in the :mod:`beartype` codebase paired with a
        related object either satisfying or violating that hint.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.door import die_if_unbearable
    from beartype.roar import (
        BeartypeConfException,
        BeartypeDecorHintNonpepException,
        BeartypeDoorHintViolation,
    )
    from beartype._util.text.utiltextrepr import represent_object
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPithUnsatisfiedMetadata)
    from pytest import raises

    # ....................{ PASS                           }....................
    # For each predefined type hint and associated metadata...
    for hint_pith_meta in iter_hints_piths_meta():
        # Type hint to be type-checked.
        hint = hint_pith_meta.hint_meta.hint

        # Beartype dataclass configuring this type-check.
        conf = hint_pith_meta.hint_meta.conf

        # Object to type-check against this type hint.
        pith = hint_pith_meta.pith

        # If this pith violates this hint...
        if isinstance(hint_pith_meta.pith_meta, HintPithUnsatisfiedMetadata):
            # print(f'hint_pith_meta: {hint_pith_meta}\n')

            # Assert this raiser raises the expected exception when passed this
            # pith and hint.
            with raises(BeartypeDoorHintViolation) as exception_info:
                die_if_unbearable(pith, hint, conf=conf)

            # Exception message raised by this wrapper function.
            exception_message = str(exception_info.value)

            # Truncated representation of this pith.
            pith_repr = represent_object(pith)

            # Assert that this message contains a truncated representation of
            # this pith.
            assert pith_repr in exception_message

            # Assert that this raiser successfully replaced the temporary
            # placeholder previously prefixing this message.
            assert 'die_if_unbearable() value ' in exception_message.lower()
            assert ' violates type hint ' in exception_message
        # Else, this raiser satisfies this hint. In this case...
        else:
            # Assert this validator raises *NO* exception when passed this pith
            # and hint.
            die_if_unbearable(pith, hint, conf=conf)

    # ....................{ FAIL                           }....................
    # Assert this tester raises the expected exception when passed an invalid
    # object as the type hint.
    with raises(BeartypeDecorHintNonpepException):
        die_if_unbearable(
            obj='Holds every future leaf and flower; the bound',
            hint=b'With which from that detested trance they leap;',
        )

    # Assert this tester raises the expected exception when passed an invalid
    # object as the beartype configuration.
    with raises(BeartypeConfException):
        die_if_unbearable(
            obj='The torpor of the year when feeble dreams',
            hint=str,
            conf='Visit the hidden buds, or dreamless sleep',
        )


# See above for @ignore_warnings() discussion.
@ignore_warnings(DeprecationWarning)
def test_door_typehint_die_if_unbearable(iter_hints_piths_meta) -> None:
    '''
    Test the :meth:`beartype.door.TypeHint.die_if_unbearable` raiser method.

    This test intentionally tests only the core functionality of this tester to
    avoid violating Don't Repeat Yourself (DRY). This tester internally defers
    to the procedural :class:`beartype.door.die_if_unbearable` tester, already
    exhaustively tested by preceding unit tests.

    Parameters
    ----------
    iter_hints_piths_meta : Callable[[], Iterable[beartype_test.a00_unit.data.hint.util.data_hintmetautil.HintPithMetadata]]
        Factory function creating and returning a generator iteratively yielding
        ``HintPithMetadata`` instances, each describing a sample type hint
        exercising an edge case in the :mod:`beartype` codebase paired with a
        related object either satisfying or violating that hint.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.door import TypeHint
    from beartype.roar import (
        BeartypeDoorHintViolation,
        BeartypeDoorNonpepException,
    )
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPithUnsatisfiedMetadata)
    from contextlib import suppress
    from pytest import raises

    # ....................{ PASS                           }....................
    # For each predefined unignorable type hint and associated metadata...
    for hint_pith_meta in iter_hints_piths_meta():
        # Type hint to be type-checked.
        hint = hint_pith_meta.hint_meta.hint

        # Beartype dataclass configuring this type-check.
        conf = hint_pith_meta.hint_meta.conf

        # Object to type-check against this type hint.
        pith = hint_pith_meta.pith

        #FIXME: Remove this suppression *AFTER* improving "TypeHint" to support
        #all currently unsupported type hints.
        with suppress(BeartypeDoorNonpepException):
            # Wrapper wrapping this type hint.
            typehint = TypeHint(hint)

            # If this pith violates this hint, assert this raiser raises the
            # expected exception when passed this pith and hint.
            if isinstance(
                hint_pith_meta.pith_meta, HintPithUnsatisfiedMetadata):
                with raises(BeartypeDoorHintViolation):
                    typehint.die_if_unbearable(pith, conf=conf)
            # Else, this pith satisfies this hint. In this case, assert this
            # raiser raises *NO* exception when passed this pith and hint.
            else:
                typehint.die_if_unbearable(pith, conf=conf)

# ....................{ TESTS ~ testers                    }....................
# See above for @ignore_warnings() discussion.
@ignore_warnings(DeprecationWarning)
def test_door_is_bearable(iter_hints_piths_meta, hints_ignorable) -> None:
    '''
    Test the :class:`beartype.door.is_bearable` tester function.

    Parameters
    ----------
    iter_hints_piths_meta : Callable[[], Iterable[beartype_test.a00_unit.data.hint.util.data_hintmetautil.HintPithMetadata]]
        Factory function creating and returning a generator iteratively yielding
        ``HintPithMetadata`` instances, each describing a sample type hint
        exercising an edge case in the :mod:`beartype` codebase paired with a
        related object either satisfying or violating that hint.
    hints_ignorable : frozenset
        Frozen set of ignorable PEP-agnostic type hints.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype.door import is_bearable
    from beartype.roar import (
        BeartypeConfException,
        BeartypeDecorHintForwardRefException,
        BeartypeDecorHintNonpepException,
    )
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPithUnsatisfiedMetadata)
    from pytest import raises
    # from pytest import raises, warns
    from warnings import simplefilter

    # ....................{ SETUP                          }....................
    # # Force pytest to temporarily allow deprecation warnings to be caught by the
    # # warns() context manager for the duration of this test. By default, pytest
    # # simply "passes through" all deprecation warnings for subsequent reporting
    # # if tests otherwise successfully pass. Deprecation warnings include:
    # # * "DeprecationWarning".
    # # * "FutureWarning".
    # # * "PendingDeprecationWarning".
    # simplefilter('always')

    # ..................{ PASS ~ ignorable                   }..................
    # Arbitrary object to be tested below.
    pith = 'The breath and blood of distant lands, for ever'

    # For each predefined ignorable type hint...
    for hint_ignorable in hints_ignorable:
        # Assert this tester returns true when passed this object and this hint.
        assert is_bearable(pith, hint_ignorable) is True

    # ..................{ PASS ~ unignorable                 }..................
    # For each predefined unignorable type hint and associated metadata...
    for hint_pith_meta in iter_hints_piths_meta():
        # Metadata describing this hint.
        hint_meta = hint_pith_meta.hint_meta

        # Type hint to be type-checked.
        hint = hint_meta.hint

        # Beartype dataclass configuring this type-check.
        conf = hint_meta.conf

        # Object to type-check against this type hint.
        pith = hint_pith_meta.pith

        # True only if this pith is expected to satisfy this hint.
        is_bearable_expected = not isinstance(
            hint_pith_meta.pith_meta, HintPithUnsatisfiedMetadata)

        # True only if this pith actually did satisfy this hint.
        is_bearable_returned = is_bearable(pith, hint, conf=conf)

        #FIXME: See test_beartype() for details, please.
        # is_bearable_returned = None  # type: ignore[assignment]
        #
        # # If this tester is expected to emit a warning for this hint...
        # if hint_meta.warning_type is not None:
        #     # Decorate that function under a context manager asserting this
        #     # decoration to emit the expected warning.
        #     with warns(hint_meta.warning_type):
        #         # Boolean returned by this tester when passed these arguments.
        #         is_bearable_returned = is_bearable(pith, hint, conf=conf)
        # # Else, this tester is expected to emit *NO* warning for this hint. In
        # # this case...
        # else:
        #     # Boolean returned by this tester when passed these arguments.
        #     is_bearable_returned = is_bearable(pith, hint, conf=conf)

        # Assert that this tester returns the expected boolean when passed this
        # pith and hint.
        assert is_bearable_returned is is_bearable_expected

    # ..................{ FAIL ~ args                        }..................
    # Assert this tester raises the expected exception when passed an invalid
    # object as the type hint.
    with raises(BeartypeDecorHintNonpepException):
        is_bearable(
            obj='Holds every future leaf and flower; the bound',
            hint=b'With which from that detested trance they leap;',
        )

    # Assert this tester raises the expected exception when passed an invalid
    # object as the beartype configuration.
    with raises(BeartypeConfException):
        is_bearable(
            obj='The torpor of the year when feeble dreams',
            hint=str,
            conf='Visit the hidden buds, or dreamless sleep',
        )

    # ..................{ FAIL ~ refs                        }..................
    class RollsItsLoudWatersToTheOceanWaves(object):
        '''
        Arbitrary class with which to test relative forward references below.
        '''

        pass

    # Assert this tester raises the expected exception when passed a relative
    # forward reference as the type hint, which this tester explicitly prohibits
    # to promote both robustness and efficiency.
    with raises(BeartypeDecorHintForwardRefException):
        is_bearable(
            obj='Breathes its swift vapours to the circling air.',
            hint='RollsItsLoudWatersToTheOceanWaves',
        )


# See above for @ignore_warnings() discussion.
@ignore_warnings(DeprecationWarning)
def test_door_typehint_is_bearable(iter_hints_piths_meta) -> None:
    '''
    Test the :meth:`beartype.door.TypeHint.is_bearable` tester method.

    This test intentionally tests only the core functionality of this tester to
    avoid violating Don't Repeat Yourself (DRY). This tester internally defers
    to the procedural :class:`beartype.door.is_bearable` tester, already
    exhaustively tested by preceding unit tests.

    Parameters
    ----------
    iter_hints_piths_meta : Callable[[], Iterable[beartype_test.a00_unit.data.hint.util.data_hintmetautil.HintPithMetadata]]
        Factory function creating and returning a generator iteratively yielding
        ``HintPithMetadata`` instances, each describing a sample type hint
        exercising an edge case in the :mod:`beartype` codebase paired with a
        related object either satisfying or violating that hint.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.door import TypeHint
    from beartype.roar import BeartypeDoorNonpepException
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPithUnsatisfiedMetadata)
    from contextlib import suppress

    # ....................{ PASS                           }....................
    # For each predefined unignorable type hint and associated metadata...
    for hint_pith_meta in iter_hints_piths_meta():
        # Type hint to be type-checked.
        hint = hint_pith_meta.hint_meta.hint

        # Beartype dataclass configuring this type-check.
        conf = hint_pith_meta.hint_meta.conf

        # Object to type-check against this type hint.
        pith = hint_pith_meta.pith

        # True only if this pith satisfies this hint.
        is_bearable_expected = not isinstance(
            hint_pith_meta.pith_meta, HintPithUnsatisfiedMetadata)

        #FIXME: Remove this suppression *AFTER* improving "TypeHint" to support
        #all currently unsupported type hints.
        with suppress(BeartypeDoorNonpepException):
            # Assert this tester returns the expected boolean when passed this
            # pith and hint.
            assert TypeHint(hint).is_bearable(pith, conf=conf) is (
                is_bearable_expected)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`647`- or :pep:`742`-compliant **type guard** (i.e., objects
created by subscripting the :obj:`typing.TypeGuard` or :obj:`typing.TypeIs` type
hint factories) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import (
    BeartypeDecorHintPep647Exception,
    BeartypeDecorHintPep742Exception,
)
from beartype.typing import (
    Optional,
    Type,
)
from beartype._data.func.datafuncarg import ARG_NAME_RETURN
from beartype._data.typing.datatypingport import Hint
from beartype._data.hint.sign.datahintsigns import (
    HintSignTypeGuard,
    HintSignTypeIs,
)

# ....................{ REDUCERS                           }....................
def reduce_hint_pep647742(
    hint: Hint,
    pith_name: Optional[str],
    exception_prefix: str,
    **kwargs
) -> Type[bool]:
    '''
    Reduce the passed :pep:`647`-compliant **old-style type guard** (i.e.,
    subscription of the :obj:`typing.TypeGuard` type hint factory) *or*
    :pep:`742`-compliant **new-style type guard** (i.e., subscription of the
    :obj:`typing.TypeIs` type hint factory) to the builtin :class:`bool` class
    as advised by both :pep:`647` and :pep:`742` when performing runtime
    type-checking if this hint annotates the return of some callable (i.e., if
    ``pith_name`` is ``"return"``) *or* raise an exception otherwise (i.e., if
    this hint annotates the return of *no* callable).

    This reducer is intentionally *not* memoized (e.g., by the
    ``@callable_cached`` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Final type hint to be reduced.
    pith_name : Optional[str]
        Either:

        * If this hint annotates a parameter of some callable, the name of that
          parameter.
        * If this hint annotates the return of some callable, ``"return"``.
        * Else, :data:`None`.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    All remaining passed arguments are silently ignored.

    Returns
    -------
    Type[bool]
        Builtin :class:`bool` class.

    Raises
    ------
    BeartypeDecorHintPep647Exception
        If this type guard does *not* annotate the return of some callable
        (i.e., if ``pith_name`` is *not* :data:`.ARG_NAME_RETURN`).
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign

    # If this type guard annotates the return of some callable, reduce this type
    # guard to the builtin "bool" class. Sadly, type guards are useless at
    # runtime and exist exclusively as a means of superficially improving the
    # computational intelligence of (...wait for it) static type-checkers.
    if pith_name == ARG_NAME_RETURN:
        return bool
    # Else, this type guard does *NOT* annotate the return of some callable.

    # Sign uniquely identifying this type guard.
    hint_sign = get_hint_pep_sign(hint)

    # Substring suffixing the exception message raised below.
    exception_message_suffix = (
        f'invalid in this type hint context (i.e., '
        f'{repr(hint)} valid only as non-nested return annotation).'
    )

    # If this is a PEP 674-compliant type guard, raise an appropriate exception.
    # Type guards are contextually valid *ONLY* as top-level return annotations.
    if hint_sign is HintSignTypeGuard:
        raise BeartypeDecorHintPep647Exception(
            f'{exception_prefix}PEP 647 type guard {repr(hint)} '
            f'{exception_message_suffix}'
        )

    # Else, this *MUST* be a PEP 742-compliant type guard (by process of
    # elimination). Raise an appropriate exception.
    assert hint_sign is HintSignTypeIs, f'{repr(hint)} not PEP 742 type guard.'
    raise BeartypeDecorHintPep742Exception(
        f'{exception_prefix}PEP 742 type guard {repr(hint)} '
        f'{exception_message_suffix}'
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- or :pep:`604`-compliant **union reducers** (i.e.,
low-level callables converting union type hints to lower-level type hints more
readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._check.metadata.hint.hintsane import (
    HINT_SANE_IGNORABLE,
    HINT_SANE_RECURSIVE,
    HintOrSane,
    HintSane,
)
from beartype._data.typing.datatypingport import (
    Hint,
    ListHints,
    TupleHints,
)
from beartype._util.hint.pep.proposal.pep484604 import make_hint_pep484604_union
from beartype._util.hint.pep.utilpepget import get_hint_pep_args

# ....................{ TESTERS                            }....................
def reduce_hint_pep484604(hint: Hint, exception_prefix: str, **kwargs) -> (
    HintOrSane):
    '''
    Reduce the passed :pep:`484`- or :pep:`604`-compliant union to the ignorable
    :data:`.HINT_SANE_IGNORABLE` singleton if this union is subscripted by one or
    more **ignorable child hints** (i.e., hints that themselves reduce to the
    ignorable :data:`.HINT_SANE_IGNORABLE` singleton) *or* preserve this union as is
    otherwise (i.e., if this union is subscripted by *no* ignorable child
    hints).

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Design
    ------
    This reducer recursively reduces all child hints subscripting this union as
    a necessary prerequisite to deciding whether one or more of these hints are
    ignorable. Since one or more of these child hints may require an uncached
    reduction in the worst case, reducing unions *also* requires an uncached
    reduction in the worst case. This is that case.

    This reducer ignores all union type hint factories subscripted by one or
    more ignorable child hints, including:

    * The :pep:`484`-compliant :obj:`typing.Optional` (e.g.,
      ``typing.Optional[object]``).
    * The :pep:`484`-compliant :obj:`typing.Union` (e.g.,
      ``typing.Union[typing.Any, bool]``).
    * All :pep:`604`-compliant new-style unions (e.g., ``bool | object``).

    Why? Because unions are only as narrow as their widest child type hints.
    Shallowly ignorable hints are ignorable exactly because they are the widest
    possible hints (e.g., :class:`object`, :data:`.HINT_SANE_IGNORABLE`), which are
    so wide as to constrain nothing and convey no meaningful semantics. A union
    of one or more shallowly ignorable child hints is thus the widest possible
    union, which is so wide as to constrain nothing and convey no meaningful
    semantics. There exist a countably infinite number of possible unions
    subscripted by one or more ignorable child hints. Ergo, these subscriptions
    *cannot* be explicitly listed in the
    :data:`beartype._data.hint.datahintrepr.HINTS_REPR_IGNORABLE_SHALLOW`
    set. Instead, these subscriptions are dynamically detected by this tester at
    runtime and thus referred to as **deeply ignorable unions.**

    Parameters
    ----------
    hint : HintPep695TypeAlias
        Union hint to be reduced.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    All remaining passed keyword parameters are passed to the parent
    :func:`beartype._check.convert._reduce.redmain.reduce_hint` function
    recursively called by this reducer.

    Returns
    -------
    HintOrSane
        Either:

        * If this union is subscripted by one or more ignorable child hints,
          :data:`.HINT_SANE_IGNORABLE`.
        * Else, this union unmodified.
    '''
    # print(f'[484/604] Detecting union {repr(hint)} ignorability...')

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype._check.convert._reduce.redmain import reduce_hint_child

    # ....................{ LOCALS                         }....................
    # Tuple of the two or more child hints subscripting this union.
    hint_childs_old = get_hint_pep_args(hint)

    # Number of these child hints.
    hint_childs_len = len(hint_childs_old)

    # If this union is subscripted by *NO* child hints, this union is ignorable.
    # In this case, reduce this union to the ignorable "HINT_SANE_IGNORABLE"
    # singleton.
    #
    # Why are unsubscripted unions ignorable? First, consider the case of the
    # unsubscripted "typing.Union" type hint factory. When unsubscripted, this
    # factory semantically expands to the implicit "Union[Any]" singleton by the
    # same argument. Since PEP 484 stipulates that a union of one type
    # semantically reduces to only that type, "Union[Any]" semantically reduces
    # to merely "Any". Despite their semantic equivalency, however, these
    # objects remain syntactically distinct with respect to object
    # identification: e.g.,
    #     >>> Union is not Union[Any]
    #     True
    #     >>> Union is not Any
    #     True
    #
    # This intentionally excludes:
    # * The "Union[Any]" and "Union[object]" singletons, since the "typing"
    #   module physically reduces:
    #   * "Union[Any]" to merely "Any" (i.e., "Union[Any] is Any"), which is
    #     already ignored by reducers elsewhere.
    #   * "Union[object]" to merely "object" (i.e., "Union[object] is
    #     object"), which is already ignored by reducers elsewhere.
    # * The "Union" singleton subscripted by one or more ignorable type hints
    #   contained in this set (e.g., "Union[Any, bool, str]"). Since there exist
    #   a countably infinite number of these subscriptions, these subscriptions
    #   are recursively detected below.
    #
    # Next, consider the case of the unsubscripted "typing.Optional" type hint
    # factory. When unsubscripted, this factory semantically expands to the
    # implicit "Optional[Any]" singleton by the same argument. Since PEP 484
    # also stipulates that all "Optional[t]" singletons semantically expand to
    # "Union[t, type(None)]" singletons for arbitrary arguments "t",
    # "Optional[Any]" semantically expands to merely "Union[Any, type(None)]".
    # Since all unions subscripted by "Any" semantically reduce to merely "Any",
    # the "Optional" singleton also reduces to merely "Any".
    #
    # This intentionally excludes "Optional[type(None)]", which the "typing"
    # module physically reduces to merely "type(None)". *shrug*
    if not hint_childs_len:
        return HINT_SANE_IGNORABLE
    # Else, this union is subscripted by one or more child hints.

    # Assert this union to be subscripted by two or more child hints.
    #
    # Note this should *ALWAYS* be the case, as:
    # * The "typing" module explicitly prohibits empty union subscription: e.g.,
    #       >>> typing.Union[]
    #       SyntaxError: invalid syntax
    #       >>> typing.Union[()]
    #       TypeError: Cannot take a Union of no types.
    # * The "typing" module reduces unions of one child hint to that hint: e.g.,
    #     >>> import typing
    #     >>> typing.Union[int]
    #     int
    assert hint_childs_len >= 2, (
        f'{exception_prefix}'
        f'PEP 484 or 604 union type hint {repr(hint)} either unsubscripted '
        f'or subscripted by only one child type hint.'
    )

    # 0-based index of the currently iterated child hint.
    hint_childs_index = 0

    # List of all sanified child hints to be returned as the reduced members of
    # this union.
    hint_childs_new_list: ListHints = []

    # Instruct the higher-level reduce_hint_child() reducer called below to
    # preserve ignorable hints reduced to a unique "HintSane" object *NOT* equal
    # to the standard "HINT_SANE_IGNORABLE" singleton but instead encapsulating
    # the "HINT_IGNORABLE" type hint and unique metadata describing that hint.
    # This enables logic below to inspect these metadata, including the
    # "hint_recursable_to_depth" dictionary required to decide whether a child
    # hint is either:
    # * Recursive (and thus not ignorable in the conventional sense) *OR*...
    # * Non-recursive (and thus ignorable in the conventional sense).
    #
    # By default, reduce_hint_child() reduces such hints to the standard
    # "HINT_SANE_IGNORABLE" singleton. Though *USUALLY* desirable, that
    # reduction destroys this unique metadata required by this decision.
    kwargs['is_hint_ignorable_preserved'] = True

    # ....................{ REDUCE                         }....................
    # Note that the low-level C-based "types.UnionType" class underlying PEP
    # 604-compliant |-style unions (e.g., "int | float") imposes no constraints
    # and is thus also semantically synonymous with the ignorable "typing.Any"
    # singleton. Nonetheless, that class *CANNOT* be instantiated from Python
    # code: e.g.,
    #     >>> import types
    #     >>> types.UnionType(int, bool)
    #     TypeError: cannot create 'types.UnionType' instances
    #
    # Likewise, that class *CANNOT* be subscripted. It follows that there exists
    # no meaningful equivalent of shallow type-checking for these unions. While
    # trivially feasible, listing "<class 'types.UnionType'>" here would only
    # prevent callers from meaningfully type-checking these unions passed as
    # valid parameters or returned as valid returns: e.g.,
    #     @beartype
    #     def muh_union_printer(muh_union: UnionType) -> None: print(muh_union)
    #
    # Ergo, we intentionally omit that class from consideration here.

    # For each child hint of this union...
    while hint_childs_index < hint_childs_len:
        # Currently visited child hint of this union.
        hint_child_insane = hint_childs_old[hint_childs_index]
        # print(f'Recursively reducing {hint} child {hint_child}...')
        # print(f'hints_overridden: {kwargs["hints_overridden"]}')

        # Sane child hint sanified from this possibly insane child hint if
        # sanifying this child hint did not generate supplementary metadata *OR*
        # that metadata otherwise (i.e., if sanifying this child hint generated
        # supplementary metadata).
        hint_child_sane = reduce_hint_child(hint_child_insane, kwargs)

        # If this child hint is ignorable, reduce this entire union to the
        # "HINT_SANE_IGNORABLE" singleton. Why? By set logic, a union
        # subscripted by one or more ignorable child hints is itself ignorable.
        if hint_child_sane is HINT_SANE_IGNORABLE:
            # print(f'Ignoring union {hint} with ignorable child {hint_child_sane}...')
            return HINT_SANE_IGNORABLE
        # Else, this child hint is unignorable.
        #
        # Else if this child hint is recursive (i.e., is a transitive parent of
        # itself previously visited by the current search), shallowly ignore
        # this child hint *WITHOUT* ignoring this entire union by simply
        # removing this child hint from this union.
        #
        # Recursive child hints are ignorable in (most) other contexts. However,
        # recursive child hints are *NOT* ignorable in the usual sense inside
        # union hints. In this context, a recursive child hint is simply a union
        # hint to be shallowly rather than deeply ignored. That is, a recursive
        # child hint of a union does *NOT* propagate its ignorability to that
        # union. That union remains unignorable regardless of whether that union
        # contains a recursive child hint.
        #
        # Consider the trivial PEP 695-compliant recursive union type alias:
        #       type RecursiveUnion = int | RecursiveUnion
        #
        # That unignorable union contains the recursive child hint
        # "RecursiveUnion" but is *NOT* ignorable. Rather, that union
        # semantically reduces to the builtin "int" type. Why? Continue reading.
        #
        # The reduce_hint_child() function called above expands recursive type
        # aliases twice: once for the original alias and a second time for the
        # recursive alias embedded in that alias. Doing so preserves data
        # structures across aliases recursively containing themselves. After
        # performing these expansions, this expanded union resembles:
        #       int | int | RecursiveUnion
        #
        # Naturally, this expanded union flattens to simply "int |
        # RecursiveUnion". Equally naturally, the "RecursiveUnion" member of
        # this union is ignorable. However, the "int" member of this union is
        # *NOT* ignorable. Ergo, this union itself is *NOT* ignorable. Instead,
        # this union is semantically equivalent to the builtin "int" type.
        elif hint_child_sane is HINT_SANE_RECURSIVE:
            pass
        # Else, this child hint is non-recursive.
        #
        # If metadata encapsulates the reduction of this child hint, reduce this
        # metadata to this possibly insane child hint. While non-ideal, the
        # remainder of the codebase is currently unequipped to handle unions of
        # PEP-noncompliant "HintSane" objects. However, even if the remainder of
        # the codebase were refactored to handle such unions, there is *NO*
        # guarantee that Python itself would allow this abuse of unions.
        # Technically, both PEP 484- and 604-compliant unions prohibit
        # PEP-noncompliant child hints. Even if Python currently allowed this,
        # there is *NO* guarantee that future releases of Python would do so.
        elif isinstance(hint_child_sane, HintSane):
            hint_childs_new_list.append(hint_child_insane)
        # Else, *NO* metadata encapsulates the reduction of this child hint.
        #
        # In this case, preserve this child hint as is.
        else:
            hint_childs_new_list.append(hint_child_sane)

        # Increment the 0-based index of the currently iterated child hint.
        hint_childs_index += 1

    # ....................{ RETURN                         }....................
    # Tuple of all sanified child hints to be returned as the reduced members of
    # this union, coerced from this list.
    hint_childs_new: TupleHints = tuple(hint_childs_new_list)

    # Possibly reduced union reconstituted from the passed union, defined as
    # either...
    hint = (
        # If the above reductions preserved the same child hints, this union
        # unmodified;
        hint
        if hint_childs_old == hint_childs_new else
        # Else, the above reductions reduced at least one of these child hints.
        # In this case, a new union reconstituted from these child hints.
        make_hint_pep484604_union(hint_childs_new)
    )

    # Return this possibly reduced union.
    return hint



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-compliant **type variable type-checkers** (i.e.,
low-level callables validating that arbitrary objects satisfy a given type
variable's associated bounds and/or constraints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep484TypeVarViolation
from beartype.typing import TypeVar
from beartype._data.typing.datatyping import (
    TypeException,
)
from beartype._data.typing.datatypingport import (
    Hint,
)
from beartype._util.cls.pep.clspep3119 import is_object_issubclassable
from beartype._util.hint.nonpep.utilnonpeptest import is_hint_nonpep_type
from beartype._util.hint.pep.proposal.pep484.pep484typevar import (
    get_hint_pep484_typevar_bounded_constraints_or_none,
    # is_hint_pep484_typevar,
)
from beartype._util.hint.pep.utilpeptest import is_hint_pep

# ....................{ TODO                               }....................
#FIXME: Generalize the die_if_hint_pep484_typevar_bound_unbearable()
#type-checker to generically validate the passed hint to be a full-blown
#*SUBHINT* of the passed type parameter. Specifically, if this type parameter is
#bounded by one or more bounded constraints, then we should validate this hint
#to be a *SUBHINT* of those constraints: e.g.,
#    class MuhClass(object): pass
#
#    # PEP 695 type alias parametrized by a type parameter bound to a
#    # subclass of the "MuhClass" type.
#    type muh_alias[T: MuhClass] = T | int
#
#    # *INVALID.* Ideally, @beartype should reject this, as "int" is
#    # *NOT* a subhint of "MuhClass".
#    def muh_func(muh_arg: muh_alias[int]) -> None: pass
#
#Doing so is complicated, however, by forward reference proxies. For obvious
#reasons, forward reference proxies are *NOT* safely resolvable at this early
#decoration time that this function is typically called at. If this hint either:
#* Is itself a forward reference proxy, ignore rather than validate this hint as
#  a subhint of these bounded constraints. Doing so is trivial by simply calling
#  "is_beartype_forwardref(hint)" here.
#* Is *NOT* itself a forward reference proxy but is transitively subscripted by
#  one or more forward reference proxies, ignore rather than validate this hint
#  as a subhint of these bounded constraints. Doing so is *EXTREMELY
#  NON-TRIVIAL.* Indeed, there's *NO* reasonable way to do so directly here.
#  Rather, we'd probably have to embrace an EAFP approach: that is, just crudely
#  try to:
#  * Detect whether this hint is a subhint of these bounded constraints.
#  * If doing so raises an exception indicative of a forward reference issue,
#    silently ignore that exception.
#
#  Of course, we're unclear what exception type that would even be. Does the
#  beartype.door.is_subhint() tester even explicitly detect forward reference
#  issues and raise an appropriate exception type? No idea. Probably *NOT*,
#  honestly. Interestingly, is_subhint() currently even fails to support
#  standard PEP 484-compliant forward references:
#      >>> is_subhint('int', int)
#      beartype.roar.BeartypeDoorNonpepException: Type hint 'int'
#      currently unsupported by "beartype.door.TypeHint".
#
#Due to these (and probably more) issues, we currently *ONLY* validate this hint
#to be a subhint of these bounded constraints.

# ....................{ RAISERS                            }....................
#FIXME: Unit test us up, please. *sigh*
def die_if_hint_pep484_typevar_bound_unbearable(
    # Mandatory parameters.
    hint: Hint,
    typevar: TypeVar,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep484TypeVarViolation,
    exception_prefix: str = '',
):
    '''
    Raise an exception unless the passed type hint satisfies the bounds and/or
    constraints of the passed :pep:`484`-compliant **type variable** (i.e.,
    :class:`typing.TypeVar` object).

    Equivalently, raise an exception if this hint violates these bounds and/or
    constraints. This raiser thus superficially type-checks this hint against
    this type variable *without* regard for any lookup table previously mapping
    this type variable to another concrete type hint.

    Parameters
    ----------
    hint : Hint
        Type hint to be validated.
    hint : TypeVar
        Type variable to validate this type hint against.
    exception_cls : Type[Exception], default: BeartypeDecorHintPep484TypeVarViolation
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintPep484TypeVarViolation`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Raises
    ------
    exception_cls
        If this hint violates this type variable's bounds and/or constraints.
    '''

    # If this hint is *NOT* an isinstanceable type after explicitly rejecting
    # beartype-specific forward reference proxies as isinstanceable types, this
    # hint *COULD* possibly be a such a proxy. If this hint is such a proxy,
    # this proxy could be unresolvable at the early time this raiser is called
    # despite otherwise being valid. To avoid raising false positives, this
    # raiser avoids the conundrum entirely by reducing to a noop. *shrug*
    if not is_hint_nonpep_type(hint=hint, is_forwardref_valid=False):
        return
    # Else, this hint is an isinstanceable type.

    # PEP-compliant type hint synthesized from all bounded constraints
    # parametrizing this type variable if any *OR* "None" otherwise (i.e., if
    # this type variable was neither bounded nor constrained).
    #
    # Note that this call is intentionally passed positional rather positional
    # keywords due to memoization.
    typevar_bound = get_hint_pep484_typevar_bounded_constraints_or_none(
        typevar, exception_prefix)
    # print(f'[{typearg}] is_object_issubclassable({typevar_bound})? ...')
    # print(f'{is_object_issubclassable(typevar_bound, False)}')

    # If...
    if (
        # This type variable was bounded or constrained *AND*...
        typevar_bound is not None and
        # These bounded constraints are PEP-noncompliant *AND*...
        #
        # PEP-compliant constraints are *NOT* safely passable to the
        # isinstance() or issubclass() testers, even if they technically are
        # isinstanceable or issubclassable. Why? Consider the "typing.Any"
        # singleton. Under newer Python versions, the "typing.Any" singleton is
        # actually defined as a subclassable type. Although effectively *NO*
        # real-world types subclass "typing.Any", literally *ALL* objects
        # (including types) satisfy the "typing.Any" type hint. Passing
        # "typing.Any" as the second variable to the issubclass() tester below
        # would thus erroneously reject (rather than silently accept) *ALL*
        # objects as unconditionally violating these bounds.
        not is_hint_pep(typevar_bound) and
        # These bounded constraints are issubclassable (i.e., an object safely
        # passable as the second variable to the issubclass() builtin) *AND*...
        #
        # Note that this function is memoized and thus permits *ONLY* positional
        # variables.
        is_object_issubclassable(
            typevar_bound,
            # Ignore unresolvable forward reference proxies (i.e.,
            # beartype-specific objects referring to user-defined external types
            # that have yet to be defined).
            False,
        ) and
        # This PEP-noncompliant isinstanceable type hint is *NOT* a subclass of
        # these bounded constraints...
        not issubclass(hint, typevar_bound)  # type: ignore[arg-type]
    ):
        # Raise a type-checking violation.
        raise BeartypeDecorHintPep484TypeVarViolation(
            message=(
                f'{exception_prefix}type hint {repr(hint)} violates '
                f'PEP 484 type variable {repr(typevar)} '
                f'bounds or constraints {repr(typevar_bound)}.'
            ),
            culprits=(hint,),
        )
    # Else, this type variable was either:
    # * Unbounded and unconstrained.
    # * Bounded or constrained by a hint that is *NOT* issubclassable.
    # * Bounded or constrained by an issubclassable object that is the
    #   superclass of this corresponding hint, which thus satisfies these
    #   bounded constraints.



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **beartype import hookable decorator-hostile decorator method
submodule** (i.e., defining callables decorated by chains of one or more
**decorator-hostile decorator method** (i.e., decorators implemented as methods
but hostile to other decorators by prohibiting other decorators from being
applied after they themselves are applied in such a chain) into which the
:mod:`beartype.beartype` decorator will be injected after the last
decorator-hostile decorator, mimicking real-world usage of the
:func:`beartype.claw.beartype_package` import hook from external callers).
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeCallHintReturnViolation
from pytest import raises

# ....................{ IMPORTS                            }....................
# Exercise all possible kinds of imports of this decorator-hostile decorator
# type.

# Import this decorator-hostile decorator type directly as an attribute from
# the nested submodule defining this type decorator via an absolute
# import-from statement.
from beartype_test.a00_unit.data.func.data_decor import (
    DecoratorHostileWrapper,
    DecoratorHostileType,

    # Import this same type under an aliased name.
    DecoratorHostileType as DecoratorHostileTypeAlias,
)

# Import the root package transitively defining this type with an absolute
# import statement.
import beartype_test

# Import the nested submodule defining this decorator with an absolute
# import-from statement.
from beartype_test.a00_unit.data.func import data_decor

# Import a nested submodule transitively defining this type with a relative
# import-from statement.
from ..... import func

# ....................{ FACTORIES                          }....................
def make_decorator_hostile_instance() -> DecoratorHostileType:
    '''
    Factory function creating and returning a new instance of the
    decorator-hostile decorator type imported above, intentionally obfuscating
    the instantiation of this type behind a function call.
    '''

    # Truly, one-liners win all lines.
    return DecoratorHostileType()

# ....................{ LOCALS                             }....................
# Validate that the import hook presumably registered by the caller implicitly
# detects instantiations of the decorator-hostile decorator type imported above,
# regardless of exactly how that type was imported.

and_like_a_rose: DecoratorHostileType = make_decorator_hostile_instance()
'''
Arbitrary instance of a decorator-hostile decorator type instantiated via an
intentionally obfuscatory factory function and thus explicitly annotated as
being an instance of that type by a :pep:`526`-compliant annotated variable
assignment.
'''


slow_breathed_melodies = DecoratorHostileType()
'''
Arbitrary instance of a decorator-hostile decorator type, directly imported as
an attribute from the submodule defining that type.
'''


in_vermeil = DecoratorHostileTypeAlias()
'''
Arbitrary instance of a decorator-hostile decorator type, directly imported as
an attribute from the submodule defining that type aliased to a different local
name.
'''


tint_and_shape = (
    beartype_test.a00_unit.data.func.data_decor.DecoratorHostileType())
'''
Arbitrary instance of a decorator-hostile decorator type, accessed via the root
package transitively defining that type.
'''


in_fragrance_soft = data_decor.DecoratorHostileType()
'''
Arbitrary instance of a decorator-hostile decorator type, accessed via the
nested submodule directly defining that type.
'''


and_coolness_to_the_eye = func.data_decor.DecoratorHostileType()
'''
Arbitrary instance of a decorator-hostile decorator type, accessed via a
nested submodule only transitively defining that type.
'''

# ....................{ CALLABLES                          }....................
# Validate that the import hook presumably registered by the caller implicitly
# injects the @beartype decorator *AFTER* the first decorator-hostile decorator
# method in all chains of callable decorations.
#
# Note that the abstract syntax tree (AST) transform performing this injection
# is rather crudely implemented and thus detects only trivial global and local
# attribute names. Since "for" loop variables are *NOT* detected, we manually
# define one callable decorated by each such instance (rather than automating
# these callable definitions in a "for" loop).

@and_like_a_rose.decorator_hostile_method
def that_inlet() -> int:
    '''
    Arbitrary function trivially violating its return type hint, decorated by a
    decorator-hostile decorator method bound to a type instantiated via an
    intentionally obfuscatory factory function and thus explicitly annotated as
    being an instance of that type by a :pep:`526`-compliant annotated variable
    assignment.
    '''

    return 'That inlet to severe magnificence'


@slow_breathed_melodies.decorator_hostile_method
def to_severe_magnificence() -> int:
    '''
    Arbitrary function trivially violating its return type hint, decorated by a
    decorator-hostile decorator method bound to a type directly imported as an
    attribute from the submodule defining that type.
    '''

    return 'And gave a roar, as if of earthly fire,'


@in_vermeil.decorator_hostile_method
def stood_full_blown() -> int:
    '''
    Arbitrary function trivially violating its return type hint, decorated by a
    decorator-hostile decorator method bound to a type directly imported as an
    attribute from the submodule defining that type aliased to a different local
    name.
    '''

    return "That scar'd away the meek ethereal Hours"


@tint_and_shape.decorator_hostile_method
def to_enter_in() -> int:
    '''
    Arbitrary function trivially violating its return type hint, decorated by a
    decorator-hostile decorator method bound to a type accessed via the root
    package transitively defining that type.
    '''

    return 'Stood full blown, for the God to enter in.'


@in_fragrance_soft.decorator_hostile_method
def he_entered() -> int:
    '''
    Arbitrary function trivially violating its return type hint, decorated by a
    decorator-hostile decorator method bound to a type accessed via the nested
    submodule directly defining that type.
    '''

    return "He enter'd, but he enter'd full of wrath;"


@and_coolness_to_the_eye.decorator_hostile_method
def full_of_wrath() -> int:
    '''
    Arbitrary function trivially violating its return type hint, decorated by a
    decorator-hostile decorator method bound to a type accessed via a nested
    submodule only transitively defining that type.
    '''

    return "His flaming robes stream'd out beyond his heels,"

# ....................{ LOCALS                             }....................
# Tuple of all decorator-hostile decorator-decorated callables defined above.
FUNCS_DECOR_HOSTILE_CHECKED = (
    that_inlet,
    to_severe_magnificence,
    stood_full_blown,
    to_enter_in,
    he_entered,
    full_of_wrath,
)

# ....................{ ASSERTS                            }....................
# For each decorator-hostile decorator-decorated callable defined above...
for func_decor_hostile_checked in FUNCS_DECOR_HOSTILE_CHECKED:
    # ....................{ PASS                           }....................
    # Assert that the import hook registered by the caller preserved the
    # uncallable object created and returned by the decorator-hostile decorator
    # decorating the callables defined above.
    assert isinstance(func_decor_hostile_checked, DecoratorHostileWrapper)

    # ....................{ FAIL                           }....................
    # Assert that the import hook registered by the caller injected the
    # @beartype decorator *AFTER* the last decorator-hostile decorator
    # decorating the callables defined above.
    with raises(BeartypeCallHintReturnViolation):
        func_decor_hostile_checked.invoke()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **type hint fixtures.**

This submodule defines session-scoped fixtures yielding tuples of various
categories of related type hints. The :func:`beartype_test.a00_unit.conftest`
submodule imports and exposes these fixtures to *all* unit tests, which may then
reference these fixtures in the standard :mod:`pytest` way (i.e., as parameters
of the same names).
'''

# ....................{ IMPORTS                            }....................
from pytest import fixture

# ....................{ FIXTURES                           }....................
@fixture(scope='session')
def hints_meta(hints_pep_meta, hints_nonpep_meta) -> (
    'Tuple[HintNonpepMetadata]'):
    '''
    Session-scoped fixture yielding a tuple of **PEP-agnostic type hint
    metadata** (i.e.,
    :class:`beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintNonpepMetadata`
    instances, each describing a sample type hint exercising an edge case in the
    :mod:`beartype` codebase -- including both PEP-compliant and -noncompliant
    type hints).

    Parameters
    ----------
    hints_pep_meta : Tuple[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        Tuple of PEP-compliant type hint metadata describing PEP-compliant type
        hints exercising edge cases in the :mod:`beartype` codebase.
    hints_nonpep_meta : Tuple[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintNonpepMetadata]
        Tuple of PEP-noncompliant type hint metadata describing PEP-noncompliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # One world. One-liner. Let's get together and code alright.
    yield hints_pep_meta + hints_nonpep_meta


@fixture(scope='session')
def hints_ignorable(
    hints_pep_ignorable_shallow,
    hints_pep_ignorable_deep,
) -> frozenset:
    '''
    Session-scoped fixture yielding a frozen set of **ignorable
    PEP-agnostic type hints,** including both PEP-compliant *and* -noncompliant
    and shallowly *and* deeply ignorable type hints.

    Parameters
    ----------
    hints_pep_ignorable_shallow : frozenset
        Frozen set of all shallowly ignorable PEP-compliant type hints.
    hints_pep_ignorable_deep : frozenset
        Frozen set of all deeply ignorable PEP-compliant type hints.
    '''

    # I code the one-liner, but I did not code the two-liner.
    yield (
        hints_pep_ignorable_shallow |
        hints_pep_ignorable_deep
    )

# ....................{ FIXTURES ~ not : sets              }....................
@fixture(scope='session')
def not_hints_nonpep(hints_pep_hashable) -> frozenset:
    '''
    Session-scoped fixture yielding a frozen set of various objects that are
    *not* PEP-noncompliant type hints exercising well-known edge cases.

    Parameters
    ----------
    hints_pep_hashable : frozenset
        Tuple of PEP-compliant type hint metadata describing PEP-compliant type
        hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # Defer fixture-specific imports.
    from beartype_test.a00_unit.data.hint.data_hint import NOT_HINTS_HASHABLE
    from beartype._cave._cavefast import NoneType

    # Yield a frozen set of...
    yield frozenset((
        # Set comprehension of tuples containing PEP-compliant type hints. Although
        # tuples containing PEP-noncompliant type hints are themselves valid
        # PEP-noncompliant type hints supported by @beartype, tuples containing
        # PEP-compliant type hints are invalid and thus unsupported.
        {
            # Tuple containing this PEP-compliant type hint...
            (int, hint_pep_hashable, NoneType,)
            # For each hashable PEP-compliant type hint...
            for hint_pep_hashable in hints_pep_hashable
            # That is neither:
            # * An isinstanceable class.
            # * A string-based forward reference.
            #
            # Both are unique edge cases supported as both PEP 484-compliant outside
            # tuples *AND* beartype-specific inside tuples. Including these hints
            # here would erroneously cause tests to treat tuples containing these
            # hints as *NOT* tuple type hints.
            if not isinstance(hint_pep_hashable, (str, type))
        } |
        # Set comprehension of hashable PEP-compliant non-class type hints.
        hints_pep_hashable |
        # Hashable objects invalid as type hints (e.g., scalars).
        NOT_HINTS_HASHABLE
    ))



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype validator unit tests.**

This submodule unit tests the private :mod:`beartype.vale._core._valecore`
submodule.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                             }....................
#FIXME: Uncomment after refactoring "BeartypeValidatorBinaryABC", please.
# def test_api_vale_validator_conjunction() -> None:
#     '''
#     Test successful usage of the private
#     :class:`beartype.vale._core._valecore.BeartypeValidatorConjunction` class.
#     '''
#
#     # Defer test-specific imports.
#     from beartype.roar import BeartypeValeSubscriptionException
#     from beartype.vale._core._valecore import BeartypeValidator
#     from beartype.vale._core._valecorebinary import (
#         BeartypeValidatorConjunction)
#     from pytest import raises
#
#     not_though_the_soldier_knew = lambda text: bool('Someone had blundered.')
#     validator_operand_1 = BeartypeValidator(
#         is_valid_code="({obj} == 'Was there a man dismayed?')",
#         is_valid=not_though_the_soldier_knew,
#         is_valid_code_locals={'yum': not_though_the_soldier_knew},
#         get_repr=lambda: "Is[lambda text: bool('Someone had blundered.')]",
#     )
#
#     # Assert that a beartype validator preserves delimited code as is.
#     validator_delimited = BeartypeValidatorConjunction(
#         validator_operand_1=validator_operand_1,
#     )
#
#     # Assert that attempting to instantiate the "BeartypeValidator" class with
#     # non-string code raises the expected exception.
#     with raises(BeartypeValeSubscriptionException):
         



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **import hookable Python optimization integration submodule** (i.e.,
data module defining various decoratable objects, which the
:mod:`beartype.beartype` decorator will then be applied to by
:func:`beartype.claw` import hooks only if the active Python interpreter is
unoptimized by either the ``-O`` command-line option *or* the
``${PYTHONOPTIMIZE}`` environment variable).
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDoorHintViolation
from beartype._util.py.utilpyinterpreter import is_python_optimized
from pytest import raises

# ....................{ PEP 526                            }....................
# This submodule intentionally tests type-checking violations *ONLY* with
# respect to PEP 526-compliant annotated variable assignments. This submodule
# does *NOT* bother testing type-checking violations with respect to either
# callables *OR* types. Why? Because prior unit tests have already validated the
# @beartype decorator to reduce to a noop when the active Python interpreter is
# optimized. The "beartype.claw" import hook type-checks callables and types by
# decorating those callables and types by the @beartype decorator. Ergo,
# re-testing callable and type decoration here would *NOT* unambiguously
# validate that import hooks reduce to a noop when the active Python interpreter
# is optimized. After all, regardless of whether import hooks correctly reduce
# to a noop, the @beartype decorator is already guaranteed to reduce to a noop.
#
# The *ONLY* means of unambiguously validating that import hooks themselves
# reduce to a noop is to type-check unique syntax *NOT* already decoratable by
# the @beartype decorator. PEP 526-compliant annotated variable assignments are
# that syntax. After all, you can't decorate an assignment. Not yet, anyway!

# If the active Python interpreter is optimized...
if is_python_optimized():
    # Implicitly assert that a similar call raises *NO* exception, exactly as
    # expected when the active Python interpreter is optimized.
    to_the_high_towers: int = "Jarr'd his own golden region; and before"
# Else, the active Python interpreter is optimized. In this case...
else:
    # Explicitly assert that the call to the beartype.door.die_if_unbearable()
    # function injected by the "beartype.claw" import hook *AFTER* this PEP
    # 526-compliant annotated variable assignment (whose assigned value violates
    # the hint annotating this assignment) raises the expected
    # @beartype-specific type-checking violation.
    with raises(BeartypeDoorHintViolation):
        and_from_the_basements_deep: int = (
            'And from the basements deep to the high towers')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) literal type hint
classes** (i.e., :class:`beartype.door.TypeHint` subclasses implementing support
for :pep:`586`-compliant :attr:`typing.Literal` type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorsuper import TypeHint
from beartype.door._cls.doorhint import TupleTypeHints

# ....................{ SUBCLASSES                         }....................
class LiteralTypeHint(TypeHint):
    '''
    **Literal type hint wrapper** (i.e., high-level object encapsulating a
    low-level :pep:`586`-compliant :attr:`typing.Literal` type hint).
    '''

    # ..................{ PRIVATE ~ properties               }..................
    @property
    def _args_wrapped_tuple(self) -> TupleTypeHints:

        # Return the empty tuple, thus presenting "Literal" type hints as having
        # *NO* child type hints. Why? Because the arguments subscripting a
        # Literal type hint are *NOT* generally PEP-compliant type hints and
        # thus *CANNOT* be safely wrapped by "TypeHint" instances. These
        # arguments are merely arbitrary values.
        #
        # Note that this property getter is intentionally *NOT* memoized with
        # @property_cached, as Python already efficiently guarantees the empty
        # tuple to be a singleton.
        return ()


    @property
    def _is_args_ignorable(self) -> bool:
        return False

    # ..................{ PRIVATE ~ testers                  }..................
    def _is_subhint(self, other: TypeHint) -> bool:

        # If the passed hint is also a literal, return true only if the set of
        # all child hints subscripting this literal is a subset of the set of
        # all child hints subscripting that literal.
        if isinstance(other, LiteralTypeHint):
            return all(self_arg in other._args for self_arg in self._args)
        # Else, the passed hint is *NOT* also a literal.

        # Return true only if either...
        return (
            # The class of each child hint subscripting this literal is a
            # subhint (e.g., subclass) of the passed hint *OR*...
            #
            # Note that, unlike most type hints, each child hints subscripting
            # this literal is typically *NOT* a valid type hint in and of itself
            # (e.g., "Literal[True]" is a valid type hint, but "True" is not).
            # This test *CANNOT* be reduced to the simpler and sensible variant:
            #     return all(
            #         hint_child.is_subhint(other)
            #         for hint_child in self._args_wrapped_tuple
            #     )
            all(
                TypeHint(type(literal_child)).is_subhint(other)  # pyright: ignore
                for literal_child in self._args
            ) or
            # Else, the class of one or more child hints subscripting this
            # literal is *NOT* a subhint (e.g., subclass) of the passed hint.
            #
            # Defer to the superclass implementation of this method. Why?
            # Because this literal could still be a subhint of passed hint
            # according to standard typing semantics. Notably, this literal
            # could be a child type hint and thus a subhint of the passed type
            # hint - despite failing all of the above literal-specific subhint
            # tests: e.g.,
            #     # The call below handles this surprisingly common edge case.
            #     >>> Literal[True] <= Union[Literal[True], Literal[False]]
            #     True
            super()._is_subhint(other)
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **object factory utilities** (i.e., low-level callables
instantiating arbitrary objects in a general-purpose manner).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilTypeException
from beartype.typing import Optional
from beartype._cave._cavemap import NoneTypeOr
from beartype._data.typing.datatyping import (
    T,
    FrozenSetStrs,
    LexicalScope,
    TypeException,
)
from beartype._data.kind.datakindiota import SENTINEL

# ....................{ PERMUTERS                          }....................
#FIXME: Unit test us up, please.
def permute_object(
    # Mandatory parameters.
    obj: T,
    init_arg_name_to_value: LexicalScope,
    init_arg_names: FrozenSetStrs,

    # Optional parameters.
    copy_var_names: Optional[FrozenSetStrs] = None,
    exception_cls: TypeException = _BeartypeUtilTypeException,
) -> T:
    '''
    Shallow copy of the passed object such that each passed keyword parameter
    overwrites an instance variable of the same name in this copy.

    This function effectively offers an intelligent alternative to the standard
    brute-force :func:`copy.copy` function, which lacks support for parameter
    permutation.

    Caveats
    -------
    **This function intentionally modifies the passed**
    ``init_arg_name_to_value`` **dictionary for efficiency.** Notably, this
    function "fills in" all missing parameters of that dictionary. For the name
    of each **missing parameter** (i.e., name in the passed ``copy_var_names``
    set but *not* in the passed ``init_arg_name_to_value`` dictionary), this
    function adds a new key-value pair to the passed ``init_arg_name_to_value``
    dictionary mapping from this name to the current value of an instance
    variable of the same name on this object.

    Parameters
    ----------
    obj : T
        Object to be permuted.
    init_arg_name_to_value : LexicalScope
        Dictionary mapping from the name to value of each:

        * Parameter to be passed to the ``__init__`` method of the type of the
          passed object.
        * Corresponding instance variable of the passed object.
    init_arg_names : FrozenSetStrs
        Frozen set of the names of *all* parameters accepted by the ``__init__``
        method of the type of the passed object.
    copy_var_names : Optional[FrozenSetStrs], default: None
        Frozen set of the names of *all* instance variables of the passed object
        whose values will be copied as the default values of all **unpassed
        parameters** (i.e., parameters in the passed ``init_arg_names`` set but
        *not* in the passed ``init_arg_name_to_value`` dictionary), defined as
        either:

        * If this set differs from that of ``init_arg_names``, this set. It is
          the caller's responsibility to ensure that **unpassed undefaulted
          parameters** (i.e., parameters neither in this ``copy_var_names`` set
          *nor* the passed ``init_arg_name_to_value`` dictionary) are declared
          as optional by the ``__init__`` method of the type of the passed
          object, which is then responsible for defaulting these parameters.
        * If these two sets are equal, :data:`None`. In this case, this
          parameter defaults to the passed ``init_arg_names`` set.

        Defaults to :data:`None` and thus ``init_arg_names``.
    exception_cls : TypeException, optional
        Type of exception to raise in the event of a fatal error. Defaults to
        :exc:`._BeartypeUtilTypeException`.

    Returns
    -------
    T
        Shallow copy of this object such that each keyword parameter overwrites
        the instance variable of the same name in this copy.

    Raises
    ------
    exception_cls
        If the name of any passed keyword parameter is either:

        * *Not* that of a parameter accepted by the :meth:`init` method of the
          type of this object.
        * *Not* that of an existing instance variable of this object.

    Examples
    --------
    .. code-block:: pycon

       >>> from beartype._util.utilobjmake import permute_object

       >>> sleuth = ViolationCause(
       ...     pith=[42,]
       ...     hint=typing.List[int],
       ...     cause_indent='',
       ...     exception_prefix='List of integers',
       ... )
       >>> sleuth_copy = permute_object(
       ...     obj=sleuth,
       ...     init_arg_name_to_value=dict(pith=[24,]),
       ...     init_arg_names=frozenset((
       ...         'hint', 'pith', 'cause_indent', 'exception_prefix',)),
       ... )

       >>> sleuth_copy.pith
       [24,]
       >>> sleuth_copy.hint
       typing.List[int]
    '''
    assert isinstance(init_arg_name_to_value, dict), (
        f'{repr(init_arg_name_to_value)} not dictionary.')
    assert isinstance(init_arg_names, frozenset), (
        f'{repr(init_arg_names)} not frozen set.')
    assert isinstance(copy_var_names, NoneTypeOr[frozenset]), (
        f'{repr(copy_var_names)} neither frozen set nor "None".')
    assert isinstance(exception_cls, type), (
        f'{repr(exception_cls)} not exception type.')

    # ....................{ LOCALS                         }....................
    # Type of this object.
    cls = obj.__class__

    # If the caller passed *NO* frozen set of the names of instance variables of
    # this object whose values will be copied as the default values of all
    # unpassed parameters, default this to the frozen set of the names of all
    # parameters accepted by the __init__() method of the type of this object.
    if copy_var_names is None:
        copy_var_names = init_arg_names
    # Else, the caller passed this frozen set. Preserve this set as is.

    # ....................{ VALIDATE                       }....................
    # For the name of each passed keyword parameter...
    for init_arg_name in init_arg_name_to_value:
        # If this name is *NOT* that of a parameter accepted by the __init__()
        # method of this type, raise an exception.
        if init_arg_name not in init_arg_names:
            raise exception_cls(
                f'{cls}.__init__() parameter "{init_arg_name}" unrecognized.')
        # Else, this name is that of a parameter accepted by that method.

    # ....................{ DEFAULT                        }....................
    # For the name of each instance variable of this object to be passed as the
    # default value of the unpassed parameter of the same name accepted by that
    # method...
    for copy_var_name in copy_var_names:
        # If this name is *NOT* that of a parameter accepted by the __init__()
        # method of this type, raise an exception.
        if copy_var_name not in init_arg_names:
            raise exception_cls(
                f'{cls}.__init__() parameter "{copy_var_name}" unrecognized.')
        # Else, this name is that of a parameter accepted by that method.

        # If this parameter was *NOT* explicitly passed by the caller...
        if copy_var_name not in init_arg_name_to_value:
            # Current value of this parameter as an instance variable of this
            # object if this object defines this variable *OR* the sentinel
            # placeholder otherwise.
            copy_var_value = getattr(obj, copy_var_name, SENTINEL)

            # If the current value of this parameter is *NOT* the sentinel
            # placeholder, this object defines this variable. In this case...
            if copy_var_value is not SENTINEL:
                # Default this parameter to its current value from this object.
                init_arg_name_to_value[copy_var_name] = copy_var_value
            # Else, this object fails to define this variable. In this case,
            # assume this parameter to be optional and thus safely undefinable.
        # Else, this parameter was explicitly passed by the caller. In this
        # case, preserve this parameter as is.

    # ....................{ RETURN                         }....................
    # New instance of this class initialized with these parameter.
    object_permuted = cls(**init_arg_name_to_value)

    # Return this instance.
    return object_permuted



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **text prefix utilities** (i.e., low-level callables creating and
returning human-readable strings describing prominent objects or types and
*always* suffixed by exactly one space character, intended to prefix
human-readable error messages).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.func.datafuncarg import ARG_NAME_RETURN
from beartype._data.typing.datatyping import (
    BeartypeableT,
    BoolTristate,
)
from beartype._util.text.utiltextlabel import (
    label_callable,
    label_type,
)
from collections.abc import Callable

# ....................{ PREFIXERS ~ object                 }....................
#FIXME: Unit test this prefixer up with respect to classes, please.
def prefix_object(obj: object, **kwargs) -> str:
    '''
    Human-readable label describing the passed arbitrary object suffixed by
    delimiting whitespace.

    Parameters
    ----------
    obj : object
        Object to be labelled.

    All remaining parameters are passed as is to the lower-level
    :func:`.label_object` function.

    Returns
    -------
    str
        Human-readable label describing this object.
    '''

    # Avoid circular import dependencies.
    from beartype._util.text.utiltextlabel import label_object

    # Return the label describing this object suffixed by delimiting whitespace.
    return f'{label_object(obj, **kwargs)} '

# ....................{ PREFIXERS : callable : name        }....................
def prefix_callable_pith(
    # Mandatory parameters.
    func: Callable,
    pith_name: str,

    # Optional parameters.
    is_color: BoolTristate = False,
) -> str:
    '''
    Human-readable label describing either the parameter with the passed name
    *or* return value if this name is ``"return"`` of the passed **beartypeable
    callable** (i.e., callable wrapped by the :func:`beartype.beartype`
    decorator with a wrapper function type-checking that callable) suffixed by
    delimiting whitespace.

    Parameters
    ----------
    func : Callable
        Decorated callable to be labelled.
    pith_name : str
        Name of the parameter or return value of this callable to be labelled.
    is_color : BoolTristate, optional
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.

    Returns
    -------
    str
        Human-readable label describing either the name of this parameter *or*
        this return value.
    '''
    assert isinstance(pith_name, str), f'{repr(pith_name)} not string.'

    # Return a human-readable label describing either...
    return (
        # If this name is "return", the return value of this callable.
        prefix_callable_return(func=func, is_color=is_color)
        if pith_name == ARG_NAME_RETURN else
        # Else, the parameter with this name of this callable.
        prefix_callable_arg_name(
            func=func, arg_name=pith_name, is_color=is_color)
    )


def prefix_callable_arg_name(
    # Mandatory parameters.
    func: Callable,
    arg_name: str,

    # Optional parameters.
    is_color: BoolTristate = False,
) -> str:
    '''
    Human-readable label describing the parameter with the passed name of the
    passed **decorated callable** (i.e., callable wrapped by the
    :func:`beartype.beartype` decorator with a wrapper function type-checking
    that callable) suffixed by delimiting whitespace.

    Parameters
    ----------
    func : Callable
        Decorated callable to be labelled.
    arg_name : str
        Name of the parameter of this callable to be labelled.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.

    Returns
    -------
    str
        Human-readable label describing this parameter's name.
    '''
    assert isinstance(arg_name, str), f'{repr(arg_name)} not string.'

    # Avoid circular import dependencies.
    from beartype._util.text.utiltextansi import color_arg_name

    # Double-quote this argument name.
    arg_name = f'"{arg_name}"'

    # Create and return this label.
    return (
        f'{prefix_object(obj=func, is_color=is_color)}'
        f'parameter {color_arg_name(text=arg_name, is_color=is_color)} '
    )


def prefix_callable_return(
    # Mandatory parameters.
    func: Callable,

    # Optional parameters.
    is_color: BoolTristate = False,
) -> str:
    '''
    Human-readable label describing the return of the passed **decorated
    callable** (i.e., callable wrapped by the :func:`beartype.beartype`
    decorator with a wrapper function type-checking that callable) suffixed by
    delimiting whitespace.

    Parameters
    ----------
    func : Callable
        Decorated callable to be labelled.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.

    Returns
    -------
    str
        Human-readable label describing this return.
    '''

    # Create and return this label.
    return f'{prefix_object(obj=func, is_color=is_color)}return '

# ....................{ PREFIXERS : callable : value       }....................
def prefix_callable_arg_value(
    # Mandatory parameters.
    func: Callable,
    arg_name: str,
    arg_value: object,

    # Optional parameters.
    is_color: BoolTristate = False,
) -> str:
    '''
    Human-readable label describing the parameter with the passed name and
    trimmed value of the passed **decorated callable** (i.e., callable wrapped
    by the :func:`beartype.beartype` decorator with a wrapper function
    type-checking that callable) suffixed by delimiting whitespace.

    Parameters
    ----------
    func : Callable
        Decorated callable to be labelled.
    arg_name : str
        Name of the parameter of this callable to be labelled.
    arg_value : object
        Value of the parameter of this callable to be labelled.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.

    Returns
    -------
    str
        Human-readable label describing this parameter's name and value.
    '''
    assert isinstance(arg_name, str), f'{repr(arg_name)} not string.'

    # Avoid circular import dependencies.
    #
    # Note that this function differs enough from the comparable
    # prefix_callable_arg_name() function to warrant a distinct implementation.
    from beartype._util.text.utiltextansi import color_arg_name

    # Create and return this label.
    return (
        f'{prefix_object(obj=func, is_color=is_color)}'
        f'parameter {color_arg_name(text=arg_name, is_color=is_color)}='
        f'{prefix_pith_value(pith=arg_value, is_color=is_color)}'
    )


def prefix_callable_return_value(
    # Mandatory parameters.
    func: Callable,
    return_value: object,

    # Optional parameters.
    is_color: BoolTristate = False,
) -> str:
    '''
    Human-readable label describing the passed trimmed return value of the
    passed **decorated callable** (i.e., callable wrapped by the
    :func:`beartype.beartype` decorator with a wrapper function type-checking
    that callable) suffixed by delimiting whitespace.

    Parameters
    ----------
    func : Callable
        Decorated callable to be labelled.
    return_value : object
        Value returned by this callable to be labelled.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.

    Returns
    -------
    str
        Human-readable label describing this return value.
    '''

    # Create and return this label.
    return (
        f'{prefix_callable_return(func=func, is_color=is_color)}'
        f'{prefix_pith_value(pith=return_value, is_color=is_color)}'
    )

# ....................{ PREFIXERS ~ pith                   }....................
def prefix_pith_type(
    # Mandatory parameters.
    pith: object,

    # Optional parameters.
    is_color: BoolTristate = False,
) -> str:
    '''
    Human-readable label describing the passed type of the **current pith**
    (i.e., arbitrary object violating the current type check) suffixed by
    delimiting whitespace.

    Parameters
    ----------
    pith : object
        Arbitrary object violating the current type check.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.

    Returns
    -------
    str
        Human-readable label describing this pith type.
    '''

    # Avoid circular import dependencies.
    from beartype._util.text.utiltextansi import color_type
    from beartype._util.text.utiltextlabel import label_object_type

    # To boldly go where no one-liner has gone before.
    return color_type(text=f'{label_object_type(pith)} ', is_color=is_color)


def prefix_pith_value(
    # Mandatory parameters.
    pith: object,

    # Optional parameters.
    is_color: BoolTristate = False,
) -> str:
    '''
    Human-readable label describing the passed value of the **current pith**
    (i.e., arbitrary object violating the current type check) suffixed by
    delimiting whitespace.

    Parameters
    ----------
    pith : object
        Arbitrary object violating the current type check.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.

    Returns
    -------
    str
        Human-readable label describing this pith value.
    '''

    # Avoid circular import dependencies.
    from beartype._util.text.utiltextlabel import label_pith_value

    # Create and return this label.
    return f'{label_pith_value(pith=pith, is_color=is_color)} '



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **beartype import hookable synchronous function submodule** (i.e.,
data module containing *only* annotated synchronous functions, mimicking
real-world usage of the :func:`beartype.claw.beartype_package` import hook from
an external caller).

See Also
--------
:mod:`beartype_test.a00_unit.data.claw.intraprocess.hookable_package.kind.data_claw_coro`
    Comparable beartype import hookable **coroutine** (i.e., asynchronous
    function) submodule. Coroutines are *only* callable from within an
    **asynchronous event loop** (e.g., :mod:`asyncio.run`) unavailable from
    within this asynchronous-agnostic submodule.
'''

# ....................{ IMPORTS                            }....................
from beartype import (
    beartype,
    BeartypeConf,
)
from beartype.roar import (
    BeartypeClawDecorWarning,
    BeartypeCallHintParamViolation,
    BeartypeCallHintReturnViolation,
    BeartypeDecorHintPep484Exception,
)
from beartype.typing import (
    NoReturn,
    Optional,
    Union,
)
from pytest import (
    raises,
    warns,
)

# from beartype.claw._importlib.clawimpcache import module_name_to_beartype_conf
# print(f'this_submodule conf: {repr(module_name_to_beartype_conf)}')

# ....................{ FUNCTIONS                          }....................
def thee_ever_and_thee_only(i_have_watched: Union[float, bytes]) -> (
    Optional[complex]):
    '''
    Arbitrary synchronous function either returning the passed float first
    doubled and then coerced into a complex number with imaginary component
    ``1`` if this float is non-zero *or* raising a
    :exc:`.BeartypeCallHintParamViolation` exception otherwise (i.e., if this
    float is zero), exercising that beartype import hooks decorate global
    synchronous functions as expected.
    '''

    def thy_shadow(and_the_darkness_of_thy_steps: Optional[float]) -> (
        Union[complex, str]):
        '''
        Arbitrary synchronous closure either returning the passed float first
        doubled and then coerced into a complex number with imaginary component
        ``1`` if this float is non-:data:`None` *or* raising a
        :exc:`.BeartypeCallHintReturnViolation` exception otherwise (i.e., if
        this float is :data:`None`), exercising that beartype import hooks
        decorate synchronous closures as expected.
        '''

        # Return either...
        return (
            # If this float is "None", explicitly force this
            # @beartype-decorated closure to raise a
            # "BeartypeCallHintReturnViolation" exception;
            None
            if and_the_darkness_of_thy_steps is None else
            # Else, this float is non-"None". In this case, return this float
            # doubled and then coerced into a complex number with imaginary
            # component "1". Why? Just because. *THIS IS BEARTYPE.* Graaaah!
            i_have_watched + and_the_darkness_of_thy_steps + 1j
        )

    # Return either...
    return (
        # If this float is zero, explicitly force this @beartype-decorated
        # closure to raise a "BeartypeCallHintReturnViolation" exception.
        thy_shadow(None)
        if i_have_watched == 0.0 else
        # Else, this integer is non-zero. In this case, return this integer
        # doubled and then coerced into a complex number with imaginary
        # component "1". Why? Just because. *THIS IS BEARTYPE.* Graaaah!
        thy_shadow(i_have_watched)
    )

# ....................{ PASS                               }....................
# Assert that calling this function passed an arbitrary float returns the
# expected complex number *WITHOUT* raising an exception.
assert thee_ever_and_thee_only(
    len('And my heart ever gazes on the depth') + 0.0) == 72 + 1j

# ....................{ FAIL                               }....................
# Assert that calling this function passed an invalid parameter raises the
# expected exception.
with raises(BeartypeCallHintParamViolation):
    thee_ever_and_thee_only(len('Of thy deep mysteries. I have made my bed'))

# Assert that calling this function passed zero raises the expected exception
# from the closure defined and called by this function.
with raises(BeartypeCallHintReturnViolation):
    thee_ever_and_thee_only(
        len('In and on coffins') -       # <-- same string length *wat*
        len('where black death') + 0.0)  # <-- same string length *lol*

# ....................{ FAIL ~ decoration                  }....................
# Assert that attempting to define a function whose type hints violate one or
# more PEP standards at decoration time emits the expected non-fatal warning.
with warns(BeartypeClawDecorWarning):
    def but_the_charmed_eddies(of_autumnal_winds: NoReturn) -> None:
        '''
        Arbitrary callable accepting a parameter erroneously annotated with the
        :pep:`484`-compliant :obj:`typing.NoReturn` type hint *only* allowed on
        returns and thus raising a decoration-time exception, which the
        :mod:`beartype.claw` API then coerces into a non-fatal warning.
        '''

        pass

# Assert that attempting to define a function whose type hints violate one or
# more PEP standards at decoration time emits the expected exception when that
# function is configured by a non-default configuration instructing the
# "beartype.claw" API to raise exceptions rather than emit non-fatal warnings at
# decoration time. *PHEW!*
with raises(BeartypeDecorHintPep484Exception):
    @beartype(conf=BeartypeConf(warning_cls_on_decorator_exception=None))
    def built_over_his_mouldering_bones(a_pyramid: NoReturn) -> None:
        '''
        Arbitrary callable accepting a parameter erroneously annotated with the
        :pep:`484`-compliant :obj:`typing.NoReturn` type hint *only* allowed on
        returns and thus raising a decoration-time exception.
        '''

        pass



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) callable type hint
classes** (i.e., :class:`beartype.door.TypeHint` subclasses implementing support
for :pep:`484`- and :pep:`585`-compliant ``Callable[...]`` type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorhint import TupleTypeHints
from beartype.door._cls.doorsuper import TypeHint
from beartype.roar import BeartypeDoorPepUnsupportedException
from beartype.typing import (
    Any,
    Tuple,
)
from beartype._data.hint.sign.datahintsignset import (
    HINT_SIGNS_CALLABLE_PARAMS)
from beartype._util.cache.utilcachecall import property_cached
from beartype._util.hint.pep.proposal.pep484585.pep484585callable import (
    get_hint_pep484585_callable_params,
    get_hint_pep484585_callable_return,
)
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

# ....................{ SUBCLASSES                         }....................
class CallableTypeHint(TypeHint):
    '''
    **Callable type hint wrapper** (i.e., high-level object encapsulating a
    low-level :pep:`484`- or :pep:`585`-compliant ``Callable[...]`` type hint).
    '''

    # ..................{ INITIALIZERS                       }..................
    def _make_args(self) -> tuple:
        # print(f'{self}._origin: {self._origin}')

        # Tuple of all child type hints subscripting this callable type hint,
        # localized for both readability and negligible efficiency gains.
        #
        # Note that this is a flattened tuple of the one or more child type
        # hints subscripting this callable type hint. Presumably for space
        # efficiency reasons, both PEP 484- *AND* 585-compliant callable type
        # hints implicitly flatten the "__args__" dunder tuple from the original
        # data structure subscripting those hints. CPython produces this
        # flattened tuple as the concatenation of:
        #
        # * Either:
        #   * If the first child type originally subscripting this hint was a
        #     list, all items subscripting the nested list of zero or more
        #     parameter type hints originally subscripting this hint as is:
        #         >>> Callable[[], bool].__args__
        #         (bool,)
        #         >>> Callable[[int, str], bool].__args__
        #         (int, str, bool)
        #
        #     This includes a list containing only the empty tuple signifying a
        #     callable accepting *NO* parameters, in which case that empty tuple
        #     is preserved as is:
        #         >>> Callable[[()], bool].__args__
        #         ((), bool)
        #   * Else, the first child type originally subscripting this hint as
        #     is. In this case, that child type is required to be either:
        #     * An ellipsis object (i.e., the "Ellipsis" builtin singleton):
        #         >>> Callable[..., bool].__args__
        #         (Ellipsis, bool)
        #     * A PEP 612-compliant parameter specification (i.e.,
        #       "typing.ParamSpec[...]" type hint):
        #         >>> Callable[ParamSpec('P'), bool].__args__
        #         (~P, bool)
        #     * A PEP 612-compliant parameter concatenation (i.e.,
        #       "typing.Concatenate[...]" type hint):
        #         >>> Callable[Concatenate[str, ParamSpec('P')], bool].__args__
        #         (typing.Concatenate[str, ~P], bool)
        # * The return type hint originally subscripting this hint.
        #
        # Note that both PEP 484- *AND* 585-compliant callable type hints
        # guarantee this tuple to contain at least one child type hint. Ergo, we
        # avoid validating that constraint here:
        #     >>> from typing import Callable
        #     >>> Callable[()]
        #     TypeError: Callable must be used as Callable[[arg, ...], result].
        #     >>> from collections.abc import Callable
        #     >>> Callable[()]
        #     TypeError: Callable must be used as Callable[[arg, ...], result].
        # args = self._args
        args = super()._make_args()

        # Note that this branch may be literally unreachable, as an
        # unsubscripted "Callable" should already be implicitly handled by the
        # "ClassTypeHint" subclass. Nonetheless, this branch exists for safety.
        if not args:  # pragma: no cover
            args = (..., Any,)
        else:
            # Parameters type hint(s) subscripting this callable type hint.
            # Note that this:
            # * May be a special object (e.g., ellipsis) rather than a tuple of
            #   zero or more parameter type hints.
            # * Has the essential side effect of eliminating harmful edge cases
            #   (e.g., "Callable[[()], Any]", which is semantically but *NOT*
            #   syntactically equivalent to "Callable[[], Any]").
            args_params = get_hint_pep484585_callable_params(self._hint)

            # Return type hint subscripting this callable type hint.
            args_return = get_hint_pep484585_callable_return(self._hint)

            # Sign uniquely identifying this parameter list if any *OR* "None".
            hint_args_sign = get_hint_pep_sign_or_none(args_params)  # pyright: ignore

            # If this hint was first subscripted by a PEP 612-compliant
            # parameter type hint, raise an exception. *sigh*
            if hint_args_sign in HINT_SIGNS_CALLABLE_PARAMS:
                raise BeartypeDoorPepUnsupportedException(
                    f'PEP 484 or 585 callable type hint {repr(self._hint)} '
                    f'PEP 612 child type hint {repr(args_params)} '
                    f'currently unsupported.'
                )
            # Else, this hint was *NOT* first subscripted by a PEP
            # 612-compliant parameter type hint.

            # Parameters type hint(s) subscripting this callable type hint,
            # coerced into a 1-tuple if *NOT* already a tuple.
            args_params_tuple = (
                args_params
                if isinstance(args_params, tuple) else
                (args_params,)
            )

            # Recreate the tuple of child type hints subscripting this parent
            # callable type hint from the tuple of argument type hints
            # introspected above. Why? Because the latter is saner than the
            # former in edge cases (e.g., ellipsis, empty argument lists).
            args = args_params_tuple + (args_return,)

        # Return these child hints.
        return args

    # ..................{ PRIVATE ~ properties               }..................
    @property
    # @property_cached
    def _args_wrapped_tuple(self) -> TupleTypeHints:

        # Tuple of all child type hints subscripting this callable type hint.
        args = self._args

        # Number of child type hints subscripting this callable type hint.
        args_len = len(args)

        # Tuple of all child type hint wrappers subscripting this callable type
        # hint wrapper, initialized to the empty tuple for simplicity.
        args_wrapped_tuple: TupleTypeHints = ()

        # If this type hint is unsubscripted, return the empty tuple.
        if not args_len:
            pass
        # Else, this type hint is subscripted by one or more child type hints.
        #
        # If this type hint is subscripted by exactly one child type hint, then
        # that child type hint signifies this callable's return type hint,
        # implying this callable accepts *NO* parameters. In this case...
        elif args_len == 1:
            # Return a 2-tuple consisting of...
            args_wrapped_tuple = (
                # Empty parameter list.
                TypeHint(Tuple[()]),  # pyright: ignore
                # Return type hint.
                TypeHint(args[-1]),
            )
        # Else, this type hint is subscripted by two or more child type hints.
        #
        # If the first child type hint subscripting this type hint is an
        # ellipsis (i.e., "..."), this callable accepts *ANY* parameters of
        # *ANY* arbitrary types. In this case...
        elif args[0] is ...:
            # Return a 2-tuple consisting of...
            args_wrapped_tuple = (
                # Variadic parameter list.
                TypeHint(Any),  # pyright: ignore
                # Return type hint.
                TypeHint(args[-1]),
            )
        # Else, the first child type hint subscripting this type hint is *NOT*
        # an ellipsis. In this case, defer to the superclass approach.
        else:
            args_wrapped_tuple = super()._args_wrapped_tuple

        # Return this tuple.
        # print(f'Callable: {self._hint}; args: {self._args}; args_wrapped_tuple: {args_wrapped_tuple}')
        return args_wrapped_tuple

    # ..................{ PROPERTIES ~ hints                 }..................
    @property  # type: ignore
    @property_cached
    def param_hints(self) -> TupleTypeHints:
        '''
        Tuple of the one or more parameter type hints subscripting this
        callable type hint.

        Notably, if this callable accepts:

        * *No* parameters (i.e., was originally subscripted by the empty list as
          ``Callable[[], ???]``), this is the 1-tuple
          ``(TypeHint(Tuple[()]),)``.
        * *Any* parameters of *any* arbitrary types (i.e., was originally
          subscripted by an ellipsis as ``Callable[..., ???]``), this is the
          1-tuple ``(TypeHint(Any),)``.
        '''

        return self._args_wrapped_tuple[:-1]


    @property
    def return_hint(self) -> TypeHint:
        '''
        Return type hint subscripting this callable type hint.
        '''

        return self._args_wrapped_tuple[-1]

    # ..................{ PROPERTIES ~ bools                 }..................
    @property
    def is_ignorable(self) -> bool:
        # Callable[..., Any] (or just `Callable`)
        return self.is_params_ignorable and self.is_return_ignorable


    @property
    def is_params_ignorable(self) -> bool:
        # Callable[..., ???]
        return self._args[0] is Ellipsis


    @property
    def is_return_ignorable(self) -> bool:
        # Callable[???, Any]
        return self.return_hint.is_ignorable

    # ..................{ PRIVATE ~ testers                  }..................
    def _is_subhint_branch(self, branch: TypeHint) -> bool:
        # print(f'Entering _is_subhint_branch({self}, {branch})...')
        # print(f'{branch}._is_args_ignorable: {branch._is_args_ignorable}')

        # If that branch is unsubscripted (e.g., "typing.Callable"), assume that
        # branch to be subscripted as the maximally wide callable type hint
        # "typing.Callable[..., Any]". Since *ALL* callable type hints are
        # necessarily subhints of that hint, return true only if the class
        # originating this hint is a subclass of the class
        # originating that branch.
        if branch._is_args_ignorable:
            return issubclass(self._origin, branch._origin)
        # Else, that branch is subscripted (e.g., "typing.Callable[..., int]").
        #
        # If that branch is *NOT* a callable type hint, this callable type hint
        # is incommensurable with that branch and thus *CANNOT* be a subhint of
        # that branch. Return false.
        elif not isinstance(branch, CallableTypeHint):
            return False
        # Else, that branch is a callable type hint.

        #FIXME: Internally comment us up, please.
        elif not branch.is_params_ignorable and (
            (
                self.is_params_ignorable or
                len(self.param_hints) != len(branch.param_hints) or
                any(
                    self_arg > branch_arg
                    for self_arg, branch_arg in zip(
                        self.param_hints, branch.param_hints)
                )
            )
        ):
            return False

        # FIXME: Insufficient, sadly. There are *MANY* different type hints that
        # are ignorable and thus semantically equivalent to "Any". It's likely
        # we should just reduce this to a one-liner resembling:
        #    return self.return_hint <= branch.return_hint
        #
        # Are we missing something? We're probably missing something. *sigh*
        elif not branch.is_return_ignorable:
            return (
                False
                if self.is_return_ignorable else
                self.return_hint <= branch.return_hint
            )

        return True



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide :pep:`585`-compliant **generics** (i.e., subclasses of a
:pep:`585`-compliant type hint factory subscripted by one or more type
parameters, exercising edge cases in unit tests requiring non-trivial generics).
'''

# ....................{ IMPORTS                            }....................
# Defer fixture-specific imports.
from beartype.typing import TypeVar
from beartype_test.a00_unit.data.data_type import Class
from beartype_test.a00_unit.data.pep.data_pep484 import (
    S,
    T,
    U,
)
from collections.abc import (
    Container as Pep585Container,
    Iterable as Pep585Iterable,
    Iterator as Pep585Iterator,
    Sequence as Pep585Sequence,
    Sized as SizedABC,
)
from contextlib import (
    AbstractContextManager as Pep585AbstractContextManager,
)

# ....................{ NON-PEP                            }....................
Nongeneric = Class
'''
Arbitrary PEP-noncompliant non-generic type.
'''

# ....................{ PEP 585 ~ non-T                    }....................
class Pep585ListStr(list[str]):
    '''
    :pep:`585`-compliant generic subclassing the builtin :class:`list` type
    subscripted (but unparametrized) by one child hint.
    '''

    # Redefine this generic's representation for debugging purposes.
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({super().__repr__()})'


class Pep585ListListStr(list[list[str]]):
    '''
    :pep:`585`-compliant generic subclassing the builtin :class:`list` type
    itself subscripted the builtin :class:`list` type subscripted (but
    unparametrized) by one child hint.
    '''

    pass

# ....................{ PEP 585 ~ T                        }....................
# Note we intentionally do *NOT* declare unsubscripted PEP 585-compliant
# generics (e.g., "class Pep585List(list):"). Why? Because PEP 585-compliant
# generics are necessarily subscripted; when unsubscripted, the corresponding
# subclasses are simply standard types.

class Pep585SequenceT(Pep585Sequence[T]):
    '''
    :pep:`585`-compliant generic sequence parametrized by one unconstrained type
    variable.
    '''

    pass

# ....................{ PEP 585 ~ T : list                 }....................
class Pep585ListT(list[T]):
    '''
    :pep:`585`-compliant generic list subclassing the builtin :class:`list` type
    parametrized by one unconstrained type variable.
    '''

    # Redefine this generic's representation for debugging purposes.
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({super().__repr__()})'


T_Pep585ListT = TypeVar('T_Pep585ListT', bound=Pep585ListT)  # <-- clever, yet sickening
'''
**Indirectly recursive type variable** (i.e., :pep:`484`-compliant type variable
inducing recursion when subscripting the otherwise non-recursive generic to
which this type variable is bound).
'''

# ....................{ PEP 585 ~ U                        }....................
class Pep585SequenceU(Pep585Sequence[U]):
    '''
    :pep:`585`-compliant generic sequence parametrized by one unconstrained type
    variable.
    '''

    pass

# ....................{ PEP 585 ~ S, T                     }....................
class Pep585DictST(dict[S, T]):
    '''
    :pep:`585`-compliant generic subclassing the builtin :class:`dict` type
    parametrized by multiple unconstrained type variables.
    '''

    # Redefine this generic's representation for debugging purposes.
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({super().__repr__()})'

# ....................{ PEP 585 ~ usable : T               }....................
# Generics that are actually instantiable and usable as valid objects.

class Pep585IterableTContainerT(Pep585Iterable[T], Pep585Container[T]):
    '''
    :pep:`585`-compliant generic subclassing multiple :mod:`collections.abc`
    abstract base classes (ABCs) directly parametrized by the same unconstrained
    type variable.
    '''

    # ................{ INITIALIZERS                           }................
    def __init__(self, sequence: tuple[T]) -> None:
        '''
        Initialize this generic from the passed tuple.
        '''

        assert isinstance(sequence, tuple), f'{repr(sequence)} not tuple.'
        self._sequence = sequence

    # ................{ ABCs                                   }................
    # Define all protocols mandated by ABCs subclassed by this generic.
    def __contains__(self, obj: object) -> bool:
        return obj in self._sequence

    def __getitem__(self, index: int) -> T:
        return self._sequence[index]

    def __iter__(self) -> Pep585Iterator[T]:
        return iter(self._sequence)

    def __len__(self) -> int:
        return len(self._sequence)


class Pep585ContextManagerTSequenceT(
    Nongeneric, Pep585AbstractContextManager[T], Pep585Sequence[T]):
    '''
    :pep:`585`-compliant generic subclassing multiple :mod:`collection.abc`
    abstract base classes (ABCs) parametrized by the same unconstrained type
    variable *and* a normal non-generic class.
    '''

    # ................{ INITIALIZERS                           }................
    def __init__(self, sequence: tuple[T]) -> None:
        '''
        Initialize this generic from the passed tuple.
        '''

        assert isinstance(sequence, tuple), f'{repr(sequence)} not tuple.'
        self._sequence = sequence

    # ................{ ABCs                                   }................
    # Define all protocols mandated by ABCs subclassed by this generic.

    def __call__(self) -> int:
        return len(self)

    def __contains__(self, obj: object) -> bool:
        return obj in self._sequence

    def __enter__(self) -> 'Pep585ContextManagerTSequenceT':
        return self

    def __exit__(self, *args, **kwargs) -> bool:
        return False

    def __getitem__(self, index: int) -> T:
        return self._sequence[index]

    def __iter__(self) -> Pep585Iterator[T]:
        return iter(self._sequence)

    def __len__(self) -> bool:
        return len(self._sequence)

    def __reversed__(self) -> tuple[T]:
        return self._sequence.reverse()

# ....................{ PEP 585 ~ usable : S, T            }....................
# Generics that are actually instantiable and usable as valid objects.

class Pep585IterableTupleSTContainerTupleST(
    SizedABC, Pep585Iterable[tuple[S, T]], Pep585Container[tuple[S, T]]):
    '''
    :pep:`585`-compliant generic subclassing multiple :mod:`collections.abc`
    abstract base classes (ABCs) indirectly parametrized (but unsubscripted) by
    multiple type variables *and* an unsubscripted and unparametrized
    :mod:`collections.abc` ABC.
    '''

    # ................{ INITIALIZERS                           }................
    def __init__(self, sequence: tuple[T]) -> None:
        '''
        Initialize this generic from the passed tuple of 2-tuples.
        '''

        assert isinstance(sequence, tuple), f'{repr(sequence)} not tuple.'
        self._sequence = sequence

    # ................{ ABCs                                   }................
    # Define all protocols mandated by ABCs subclassed by this generic.
    def __contains__(self, obj: object) -> bool:
        return obj in self._sequence

    def __iter__(self) -> bool:
        return iter(self._sequence)

    def __len__(self) -> bool:
        return len(self._sequence)

# ....................{ PEP 585 ~ usable : S, T, U         }....................
# Generics that are actually instantiable and usable as valid objects.

class Pep585ListRootU(list[U]):
    '''
    :pep:`585`-compliant generic list subclassing the builtin :class:`list` type
    parametrized by one unconstrained type variable.
    '''

    pass


class Pep585ListStemT(Pep585ListRootU[T]):
    '''
    :pep:`585`-compliant generic parametrized by one unconstrained type variable
    subclassing a :pep:`585`-compliant generic list subclassing the builtin
    :class:`list` type parametrized by a different unconstrained type variable.
    '''

    pass


class Pep585ListLeafS(Pep585ListStemT[S]):
    '''
    :pep:`585`-compliant generic parametrized by one unconstrained type variable
    subclassing a similar :pep:`585`-compliant generic subclassing a
    :pep:`585`-compliant generic list subclassing the builtin :class:`list` type
    parametrized by a different unconstrained type variable.
    '''

    pass



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype decorator :pep:`484`-compliant type hint unit tests.

This submodule unit tests the :func:`beartype.beartype` decorator with respect
to :pep:`484`-compliant type hints.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ decor : no_type_check      }....................
def test_decor_pep484_no_type_check() -> None:
    '''
    Test the :func:`beartype.beartype` decorator against all edge cases of the
    :pep:`484`-compliant :attr:`typing.no_type_check` decorator.
    '''

    # Defer test-specific imports.
    from beartype import beartype
    from beartype.typing import (
        Union,
        no_type_check,
    )

    # Callable decorated by @typing.no_type_check whose otherwise PEP-compliant
    # type hints *SHOULD* be subsequently ignored by @beartype.
    @no_type_check
    def of_beechen_green(and_shadows_numberless: Union[int, str]) -> str:
        return and_shadows_numberless

    # The same callable additionally decorated by @beartype.
    of_beechen_green_beartyped = beartype(of_beechen_green)

    # Assert these two callables to be the same, implying @beartype silently
    # reduced to a noop by returning this callable undecorated.
    assert of_beechen_green is of_beechen_green_beartyped

# ....................{ TESTS ~ hint : noreturn            }....................
def test_decor_pep484_hint_noreturn() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on synchronous callables
    against all edge cases of the :pep:`484`-compliant :attr:`typing.NoReturn`
    type hint, which is valid *only* as an unsubscripted return annotation.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import (
        BeartypeCallHintViolation,
        BeartypeDecorHintPep484Exception,
    )
    from beartype.typing import (
        NoReturn,
        Union,
    )
    from beartype_test._util.pytroar import raises_uncached

    # ..................{ LOCALS                             }..................
    # Exception guaranteed to be raised *ONLY* by the mending_wall() function.
    class BeforeIBuiltAWallIdAskToKnow(Exception): pass

    # Synchronous callable unconditionally raising an exception correctly
    # annotating its return as "NoReturn".
    @beartype
    def mending_wall() -> NoReturn:
        raise BeforeIBuiltAWallIdAskToKnow(
            "Something there is that doesn't love a wall,")

    # Callable explicitly returning a value incorrectly annotating its return
    # as "NoReturn".
    @beartype
    def frozen_ground_swell() -> NoReturn:
        return 'That sends the frozen-ground-swell under it,'

    # Callable implicitly returning a value incorrectly annotating its return
    # as "NoReturn".
    @beartype
    def we_do_not_need_the_wall() -> NoReturn:
        'There where it is we do not need the wall:'

    # ..................{ PASS                               }..................
    # Assert this callable raises the expected exception when called.
    with raises_uncached(BeforeIBuiltAWallIdAskToKnow):
        mending_wall()

    # Assert this callable raises the expected exception when called.
    with raises_uncached(BeartypeCallHintViolation):
        frozen_ground_swell()

    # Assert this callable raises the expected exception when called.
    with raises_uncached(BeartypeCallHintViolation):
        we_do_not_need_the_wall()

    # ..................{ FAIL                               }..................
    # Assert this decorator raises the expected exception when decorating a
    # synchronous callable returning a value incorrectly annotating its return
    # as "NoReturn".
    with raises_uncached(BeartypeDecorHintPep484Exception):
        @beartype
        def upper_boulders(in_the_sun: NoReturn):
            return 'And spills the upper boulders in the sun;'

    # Assert this decorator raises the expected exception when decorating a
    # synchronous callable returning a value annotating a parameter as a
    # supported PEP 484-compliant type hint incorrectly subscripted by
    # "NoReturn".
    with raises_uncached(BeartypeDecorHintPep484Exception):
        @beartype
        def makes_gaps(abreast: Union[str, NoReturn]):
            return 'And makes gaps even two can pass abreast.'


async def test_decor_pep484_hint_noreturn_async() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on asynchronous callables
    against all edge cases of the :pep:`484`-compliant :attr:`typing.NoReturn`
    type hint, which is valid *only* as an unsubscripted return annotation.
    '''

    # Defer test-specific imports.
    from beartype import beartype
    from beartype.typing import NoReturn

    # Asynchronous coroutine unconditionally raising an exception correctly
    # annotating its return as "NoReturn".
    @beartype
    async def work_of_hunters(another_thing) -> NoReturn:
        raise ValueError('The work of hunters is another thing:')

# ....................{ TESTS ~ hint : sequence            }....................
def test_decor_pep484_namedtuple() -> None:
    '''
    Test the :func:`beartype.beartype` decorator against all edge cases of
    instances of user-defined subclasses of the :pep:`484`-compliant
    :attr:`typing.NamedTuple` superclass, which are instances rather than types
    and thus invalid as actual type hints.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from typing import (
        NamedTuple,
        Optional,
    )
    from pytest import raises

    # ..................{ LOCALS                             }..................
    @beartype
    class WindsContend(NamedTuple):
        '''
        Arbitrary named tuple type-checked by :func:`beartype.beartype`.
        '''

        starbeams: Optional[str]
        '''
        Standard instance variable passed by the :class:`NamedTuple` metaclass
        to the the implicit ``__new__()`` method synthesized for this subclass.
        '''

    # Arbitrary instance of this named tuple exercising all edge cases.
    dart_through_them = WindsContend(
        starbeams='Or the star-beams dart through them. Winds contend')

    # ..................{ PASS                               }..................
    # Assert this dataclass defines the expected attributes.
    assert dart_through_them.starbeams == (
        'Or the star-beams dart through them. Winds contend')

    # ..................{ FAIL                               }..................
    # Assert that attempting to instantiate an instance of this dataclass with a
    # parameter violating the corresponding type hint annotating the field of
    # the same name raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        WindsContend(starbeams=0xBABECAFE)


def test_decor_pep484_hint_sequence_args_1_cached() -> None:
    '''
    Test that a `subtle issue <issue #5_>`__ of the :func:`beartype.beartype`
    decorator with respect to metadata describing **PEP-compliant standard
    sequence hints** (e.g., :attr:`typing.List`) cached via memoization across
    calls to that decorator has been resolved and *not* regressed.

    Note that more general-purpose :pep:`484` unit tests *should* already
    exercise this issue, but that this issue was sufficiently dire to warrant
    special-purposed testing exercising this exact issue.

    .. _issue #5:
       https://github.com/beartype/beartype/issues/5
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.typing import Union

    # ....................{ CALLABLES                      }....................
    # Callable annotated by an arbitrary PEP 484 standard sequence type hint.
    @beartype
    def fern_hill(prince_of_the_apple_towns: Union[int, str]) -> str:
        return prince_of_the_apple_towns

    # A different callable annotated by the same hint and another arbitrary
    # non-"typing" type hint.
    @beartype
    def apple_boughs(
        famous_among_the_barns: Union[int, str],
        first_spinning_place: str
    ) -> str:
        return famous_among_the_barns + first_spinning_place

    # ....................{ PASS                           }....................
    # Validate that these callables behave as expected.
    assert fern_hill(
        'Now as I was young and easy under the apple boughs'
        'About the lilting house and happy as the grass was green,'
        '  The night above the dingle starry,'
        '    Time let me hail and climb'
        '  Golden in the heydays of his eyes,'
        'And honoured among wagons I was prince of the apple towns'
        'And once below a time I lordly had the trees and leaves'
        '    Trail with daisies and barley'
        '  Down the rivers of the windfall light. '
    ).startswith('Now as I was young and easy under the apple boughs')
    assert apple_boughs((
        'And as I was green and carefree, famous among the barns'
        'About the happy yard and singing as the farm was home,'
        '  In the sun that is young once only,'
        '    Time let me play and be'
        '  Golden in the mercy of his means,'
        'And green and golden I was huntsman and herdsman, the calves'
        'Sang to my horn, the foxes on the hills barked clear and cold,'
        '    And the sabbath rang slowly'
        '  In the pebbles of the holy streams.'
    ), 'All the sun long it was running, it was lovely, the hay').startswith(
        'And as I was green and carefree, famous among the barns')

# ....................{ TESTS ~ hint : invalid             }....................
#FIXME: Excise us up. We're unconvinced that arbitrarily prohibiting *ANY*
#isinstanceable classes from being used as type hints is a useful approach --
#even if those classes are "typing" attributes typically intended to be
#instantiated when used as type hints. Nobody's here to tell anybody they can't
#do something that otherwise appears sensible.
#
#When we next read this comment, refactor our test suite as follows:
#* Remove this test.
#* Remove the
#  "beartype_test.a00_unit.data.hint.pep.data_pepfixture.HINTS_PEP_INVALID_TYPE_NONGENERIC"
#  frozen set entirely.

# def test_pep484_hint_invalid_types_nongeneric() -> None:
#     '''
#     Test the :func:`beartype.beartype` decorator against **invalid non-generic
#     classes** (i.e., classes declared by the :mod:`typing` module used to
#     instantiate PEP-compliant type hints but themselves invalid as
#     PEP-compliant type hints).
#     '''
#
#     # Defer test-specific imports.
#     from beartype import beartype
#     from beartype.roar import BeartypeDecorHintPepSignException
#     from beartype_test.a00_unit.data.hint.pep.data_pepfixture import (
#         HINTS_PEP_INVALID_TYPE_NONGENERIC)
#
#     # Assert that decorating a callable annotated by a non-generic class raises
#     # the expected exception.
#     for type_nongeneric in HINTS_PEP_INVALID_TYPE_NONGENERIC:
#         with raises_uncached(BeartypeDecorHintPepSignException):
#             @beartype
#             def childe_roland(to_the_dark_tower_came: type_nongeneric) -> (
#                 type_nongeneric):
#                 raise to_the_dark_tower_came



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`647`-compliant **type guard** (i.e., objects created by
subscripting the :obj:`typing.TypeGuard` type hint factory) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep647Exception
from beartype.typing import (
    Optional,
    Type,
)
from beartype._data.typing.datatypingport import Hint
from beartype._data.func.datafuncarg import ARG_NAME_RETURN

# ....................{ REDUCERS                           }....................
#FIXME: Unit test us up, please.
def reduce_hint_pep647(
    hint: Hint,
    pith_name: Optional[str],
    exception_prefix: str,
    **kwargs
) -> Type[bool]:
    '''
    Reduce the passed :pep:`647`-compliant **type guard** (i.e.,
    subscription of the :obj:`typing.TypeGuard` type hint factory) to the
    builtin :class:`bool` class as advised by :pep:`647` when performing
    runtime type-checking if this hint annotates the return of some callable
    (i.e., if ``pith_name`` is ``"return"``) *or* raise an exception otherwise
    (i.e., if this hint annotates the return of *no* callable).

    This reducer is intentionally *not* memoized (e.g., by the
    ``@callable_cached`` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Final type hint to be reduced.
    pith_name : Optional[str]
        Either:

        * If this hint annotates a parameter of some callable, the name of that
          parameter.
        * If this hint annotates the return of some callable, ``"return"``.
        * Else, :data:`None`.
    exception_prefix : str
        Human-readable label prefixing the representation of this object in the
        exception message.

    All remaining passed arguments are silently ignored.

    Returns
    -------
    Type[bool]
        Builtin :class:`bool` class.

    Raises
    ------
    BeartypeDecorHintPep647Exception
        If this type guard does *not* annotate the return of some callable
        (i.e., if ``arg_kind`` is *not* :data:`True`).
    '''

    # If this type guard annotates the return of some callable, reduce this type
    # guard to the builtin "bool" class. Sadly, type guards are useless at
    # runtime and exist exclusively as a means of superficially improving the
    # computational intelligence of (...wait for it) static type-checkers.
    if pith_name == ARG_NAME_RETURN:
        return bool
    # Else, this type guard does *NOT* annotate the return of some callable.

    # Raise an exception. Type guards are contextually valid *ONLY* as top-level
    # return annotations.
    raise BeartypeDecorHintPep647Exception(
        f'{exception_prefix}PEP 647 type hint {repr(hint)} '
        f'invalid in this type hint context (i.e., '
        f'{repr(hint)} valid only as non-nested return annotation).'
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Unmemoized beartype builtin descriptor decorators** (i.e., low-level
decorators decorating low-level C-based objects produced by **builtin
decorators** (i.e., :class:`classmethod`, :class:`property`,
:class:`staticmethod`)).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._cave._cavefast import (
    MethodBoundInstanceOrClassType,
    MethodDecoratorPropertyType,
)
from beartype._data.typing.datatyping import BeartypeableT
from beartype._util.func.utilfuncget import get_func_boundmethod_self
from beartype._util.func.utilfunctest import is_func_boundmethod
from beartype._util.func.utilfuncwrap import (
    unwrap_func_boundmethod_once,
    unwrap_func_class_or_static_method_once,
)

# ....................{ DECORATORS                         }....................
def beartype_descriptor_boundmethod(
    descriptor: BeartypeableT, **kwargs) -> BeartypeableT:
    '''
    Decorate the passed **builtin bound method object** (i.e., C-based bound
    method descriptor produced by Python on instantiation for each instance and
    class method defined by the class being instantiated) with dynamically
    generated type-checking.

    Parameters
    ----------
    descriptor : BeartypeableT
        Descriptor to be decorated by :func:`beartype.beartype`.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`.beartype_func` decorator internally called by this higher-level
    decorator on the pure-Python function encapsulated in this descriptor.

    Returns
    -------
    BeartypeableT
        New pure-Python callable wrapping this descriptor with type-checking.
    '''
    assert is_func_boundmethod(descriptor), (
        f'{repr(descriptor)} not builtin bound method descriptor.')

    # Avoid circular import dependencies.
    from beartype._decor._nontype.decornontype import beartype_func

    # Possibly C-based callable wrappee object encapsulated by this descriptor.
    descriptor_wrappee = unwrap_func_boundmethod_once(descriptor)

    # Instance object to which this descriptor was bound at instantiation time.
    descriptor_self = get_func_boundmethod_self(descriptor)

    # Pure-Python unbound function decorating the similarly pure-Python unbound
    # function encapsulated by this descriptor with type-checking.
    #
    # Note that doing so:
    # * Implicitly propagates dunder attributes (e.g., "__annotations__",
    #   "__doc__") from the original function onto this new function. Good.
    # * Does *NOT* implicitly propagate the same dunder attributes from the
    #   original descriptor encapsulating the original function to the new
    #   descriptor (created below) encapsulating this wrapper function. Bad!
    #   Thankfully, only one such attribute exists as of this time: "__doc__".
    #   We propagate this attribute manually below.
    func_checked = beartype_func(func=descriptor_wrappee, **kwargs)  # pyright: ignore

    # New instance method descriptor rebinding this function to the instance of
    # the class bound to the prior descriptor.
    #
    # Note that:
    # * This is required, as the "__func__" attribute of method descriptors is
    #   read-only. Attempting to do so raises this non-human-readable exception:
    #     AttributeError: readonly attribute
    #   This implies that the passed descriptor *CANNOT* be meaningfully
    #   modified. Our only recourse is to define an entirely new descriptor,
    #   effectively discarding the passed descriptor, which will then be
    #   subsequently garbage-collected. This is wasteful. This is Python.
    # * This can also be implemented by abusing the descriptor protocol:
    #       descriptor_new = descriptor_func_new.__get__(descriptor.__self__)
    #   That said, there exist *NO* benefits to doing so. Indeed, doing so only
    #   reduces the legibility and maintainability of this operation.
    descriptor_new = MethodBoundInstanceOrClassType(
        func_checked, descriptor_self)  # type: ignore[return-value]

    #FIXME: Actually, Python doesn't appear to support this at the moment.
    #Attempting to do so raises this exception:
    #    AttributeError: attribute '__doc__' of 'method' objects is not writable
    #
    #See also this open issue on the Python bug tracker requesting this be
    #resolved. Sadly, Python has yet to resolve this:
    #    https://bugs.python.org/issue47153
    # # Propagate the docstring from the prior to the new descriptor.
    # #
    # # Note that Python guarantees this attribute to exist. If the original
    # # function had a docstring, this attribute is non-"None"; else, this
    # # attribute is "None". In either case, this attribute exists. Ergo,
    # # additional validation is neither required nor desired.
    # descriptor_new.__doc__ = descriptor.__doc__

    # Return this new descriptor, implicitly destroying the prior descriptor.
    return descriptor_new  # type: ignore[return-value]


def beartype_descriptor_decorator_builtin_property(
    descriptor: BeartypeableT, **kwargs) -> BeartypeableT:
    '''
    Decorate the passed **builtin property decorator object** (i.e., C-based
    unbound method descriptor instantiated by the builtin :class:`property`
    decorator type) with dynamically generated type-checking.

    Parameters
    ----------
    descriptor : BeartypeableT
        Property descriptor to be decorated by :func:`beartype.beartype`.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`.beartype_func` decorator internally called by this higher-level
    decorator on the pure-Python function encapsulated in this descriptor.

    Returns
    -------
    BeartypeableT
        New pure-Python callable wrapping this descriptor with type-checking.
    '''
    assert isinstance(descriptor, MethodDecoratorPropertyType), (
        f'{repr(descriptor)} not builtin @property method descriptor.')

    # Avoid circular import dependencies.
    from beartype._decor._nontype.decornontype import beartype_func

    # Pure-Python unbound getter, setter, and deleter functions wrapped by this
    # descriptor if any *OR* "None" otherwise (i.e., for each such function
    # currently unwrapped by this descriptor).
    descriptor_getter  = descriptor.fget  # type: ignore[assignment,union-attr]
    descriptor_setter  = descriptor.fset  # type: ignore[assignment,union-attr]
    descriptor_deleter = descriptor.fdel  # type: ignore[assignment,union-attr]

    # Decorate this getter function with type-checking.
    #
    # Note that *ALL* property method descriptors wrap at least a getter
    # function (but *NOT* necessarily a setter or deleter function). This
    # function is thus guaranteed to be non-"None".
    descriptor_getter = beartype_func(  # type: ignore[type-var]
        func=descriptor_getter,  # pyright: ignore
        **kwargs
    )

    # If this property method descriptor additionally wraps a setter and/or
    # deleter function, type-check those functions as well.
    if descriptor_setter is not None:
        descriptor_setter = beartype_func(descriptor_setter, **kwargs)
    if descriptor_deleter is not None:
        descriptor_deleter = beartype_func(descriptor_deleter, **kwargs)

    # Return a new property method descriptor decorating all of these functions,
    # implicitly destroying the prior descriptor.
    #
    # Note that the "property" class interestingly has this signature:
    #     class property(fget=None, fset=None, fdel=None, doc=None): ...
    return property(  # type: ignore[return-value]
        fget=descriptor_getter,
        fset=descriptor_setter,
        fdel=descriptor_deleter,
        doc=descriptor.__doc__,
    )


def beartype_descriptor_decorator_builtin_class_or_static_method(
    descriptor: BeartypeableT, **kwargs) -> BeartypeableT:
    '''
    Decorate the passed **builtin class or static method decorator object**
    (i.e., C-based unbound method descriptor instantiated by either the builtin
    :class:`classmethod` or :class:`staticmethod` decorator types) with
    dynamically generated type-checking.

    Parameters
    ----------
    descriptor : BeartypeableT
        Class or static method descriptor to be decorated by
        :func:`beartype.beartype`.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`.beartype_func` decorator internally called by this higher-level
    decorator on the pure-Python function encapsulated in this descriptor.

    Returns
    -------
    BeartypeableT
        New pure-Python callable wrapping this descriptor with type-checking.
    '''

    # Avoid circular import dependencies.
    from beartype._decor.decorcore import beartype_object

    # Possibly C-based callable wrappee object decorated by this descriptor.
    #
    # Note that this wrappee is typically but *NOT* necessarily a pure-Python
    # unbound function. This descriptor explicitly permits the decorated object
    # to be a callable C-based type (i.e., defining the __call__() dunder
    # method), which numerous standard and third-party pure-Python classes then
    # leverage to augment those classes into subscriptable type hint factories
    # via a simple one-liner: e.g.,
    #     from abc import ABCMeta
    #     from beartype import beartype
    #     from types import GenericAlias
    #
    #     @beartype
    #     class MuhTypeHintFactory(metaclass=ABCMeta):
    #         # This exact one liner appears verbatim throughout the
    #         # standard library (as well as third-party packages).
    #         __class_getitem__ = classmethod(GenericAlias)
    #
    # Ergo, the name "__func__" of this dunder attribute is disingenuous. This
    # descriptor does *NOT* merely decorate functions; this descriptor
    # permissively decorates all callable objects.
    descriptor_wrappee = unwrap_func_class_or_static_method_once(descriptor)  # type: ignore[arg-type]

    # Pure-Python unbound function type-checking this wrappee. Note that:
    # * Python 3.8, 3.9, and 3.10 explicitly permit the @classmethod decorator
    #   to be chained into the @property decorator: e.g.,
    #       class MuhClass(object):
    #           @classmethod  # <-- this is fine under Python < 3.11
    #           @property
    #           def muh_property(self) -> ...: ...
    # * Python ≥ 3.11 explicitly prohibits that by emitting a non-fatal
    #   "DeprecationWarning" on each attempt to do so. Under Python ≥ 3.11,
    #   users *MUST* instead refactor the above simplistic decorator chaining
    #   use case as follows:
    #   * Define a metaclass for each class requiring a class property.
    #   * Define each class property on that metaclass rather than on that class
    #     instead.
    #
    #   In other words:
    #       class MuhClassMeta(type):  # <-- Python ≥ 3.11 demands sacrifice
    #          '''
    #          Metaclass of the :class`.MuhClass` class, defining class
    #          properties for that class.
    #          '''
    #
    #          @property
    #          def muh_property(cls) -> ...: ...
    #
    #      class MuhClass(object, metaclass=MuhClassMeta):
    #          pass
    # * Technically, all Python versions currently supported by @beartype permit
    #   this. Ergo, @beartype currently defers to:
    #   * The high-level beartype_object() decorator (which permits the passed
    #     object to be the descriptor created and returned by the @property
    #     decorator and thus implicitly allows @classmethod to be chained into
    #     @property) rather than...
    #   * The low-level beartype_func() decorator (which requires the passed
    #     object to be callable, which the descriptor created and returned by
    #     the @property decorator is *NOT*).
    descriptor_wrappee_checked = beartype_object(descriptor_wrappee, **kwargs) # type: ignore[union-attr]

    # Return a new class or static method descriptor decorating the pure-Python
    # unbound function wrapped by this descriptor with type-checking, implicitly
    # destroying the prior descriptor.
    return descriptor.__class__(descriptor_wrappee_checked)  # type: ignore[misc,return-value]
